from tamr_unify_client.base_resource import BaseResource
from tamr_unify_client.dataset.use import DatasetUse


class DatasetUsage(BaseResource):
    """
    The usage of a dataset and its downstream dependencies.

    See https://docs.tamr.com/reference#retrieve-downstream-dataset-usage
    """

    @classmethod
    def from_json(cls, client, resource_json, api_path):
        return super().from_data(client, resource_json, api_path)

    @property
    def relative_id(self):
        """:type: str"""
        return self.api_path

    @property
    def usage(self):
        """:type: :class:`~tamr_unify_client.dataset.use.DatasetUse`"""
        return DatasetUse(self.client, self._data.get("usage"))

    @property
    def dependencies(self):
        """:type: list[:class:`~tamr_unify_client.dataset.use.DatasetUse`]"""
        deps = self._data.get("dependencies")
        return [DatasetUse(self.client, dep) for dep in deps]

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"alias={self.api_path!r})"
        )



from tamr_client.dataset import dataframe, record, unified
from tamr_client.dataset._dataset import (
    _materialize_async,
    AlreadyExists,
    Ambiguous,
    attributes,
    by_name,
    by_resource_id,
    create,
    delete,
    get_all,
    materialize,
    NotFound,
)



from copy import deepcopy

from tamr_unify_client.base_resource import BaseResource


class Category(BaseResource):
    """A category of a taxonomy"""

    @classmethod
    def from_json(cls, client, data, api_path=None):
        return super().from_data(client, data, api_path)

    @property
    def name(self):
        """:type: str"""
        return self._data.get("name")

    @property
    def description(self):
        """:type: str"""
        return self._data.get("description")

    @property
    def path(self):
        """:type: list[str]"""
        return self._data.get("path")[:]

    def parent(self):
        """Gets the parent Category of this one, or None if it is a tier 1 category

        :returns: The parent Category or None
        :rtype: :class:`~tamr_unify_client.categorization.category.resource.Category`
        """
        parent = self._data.get("parent")
        if parent:
            alias = self.api_path.rsplit("/", 1)[0] + "/" + parent.split("/")[-1]
            resource_json = self.client.get(alias).successful().json()
            return Category.from_json(self.client, resource_json, alias)
        else:
            return None

    def spec(self):
        """Returns this category's spec.

        :return: The spec for the category.
        :rtype: :class:`~tamr_unify_client.categorization.category.resource.CategorySpec`
        """
        return CategorySpec.of(self)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"relative_id={self.relative_id!r}, "
            f"name={self.name!r},"
            f"path={'/'.join(self.path)!r},"
            f"description={self.description!r})"
        )


class CategorySpec:
    """A representation of the server view of a category."""

    def __init__(self, client, data, api_path):
        self.client = client
        self._data = data
        self.api_path = api_path

    @staticmethod
    def of(resource):
        """Creates a category spec from a category.

        :param resource: The existing category.
        :type resource: :class:`~tamr_unify_client.categorization.category.resource.Category`
        :return: The corresponding category spec.
        :rtype: :class:`~tamr_unify_client.categorization.category.resource.CategorySpec`
        """
        return CategorySpec(
            resource.client, deepcopy(resource._data), resource.api_path
        )

    @staticmethod
    def new():
        """Creates a blank spec that could be used to construct a new category.

        :return: The empty spec.
        :rtype: :class:`~tamr_unify_client.categorization.category.resource.CategorySpec`
        """
        return CategorySpec(None, {}, None)

    def from_data(self, data):
        """Creates a spec with the same client and API path as this one, but new data.

        :param data: The data for the new spec.
        :type data: dict
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.categorization.category.resource.CategorySpec`
        """
        return CategorySpec(self.client, data, self.api_path)

    def to_dict(self):
        """Returns a version of this spec that conforms to the API representation.

        :returns: The spec's dict.
        :rtype: dict
        """
        return deepcopy(self._data)

    def with_name(self, new_name):
        """Creates a new spec with the same properties, updating name.

        :param new_name: The new name.
        :type new_name: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.categorization.category.resource.CategorySpec`
        """
        return self.from_data({**self._data, "name": new_name})

    def with_description(self, new_description):
        """Creates a new spec with the same properties, updating description.

        :param new_description: The new description.
        :type new_description: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.categorization.category.resource.CategorySpec`
        """
        return self.from_data({**self._data, "description": new_description})

    def with_path(self, new_path):
        """Creates a new spec with the same properties, updating path.

        :param new_path: The new path.
        :type new_path: list[str]
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.categorization.category.resource.CategorySpec`
        """
        return self.from_data({**self._data, "path": new_path})

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"dict={self._data})"
        )



import json

from tamr_unify_client.base_model import MachineLearningModel
from tamr_unify_client.dataset.resource import Dataset
from tamr_unify_client.mastering.binning_model import BinningModel
from tamr_unify_client.mastering.estimated_pair_counts import EstimatedPairCounts
from tamr_unify_client.mastering.published_cluster.configuration import (
    PublishedClustersConfiguration,
)
from tamr_unify_client.mastering.published_cluster.record import RecordPublishedCluster
from tamr_unify_client.mastering.published_cluster.resource import PublishedCluster
from tamr_unify_client.project.resource import Project


class MasteringProject(Project):
    """A Mastering project in Tamr."""

    def pairs(self):
        """Record pairs generated by Tamr's binning model.
        Pairs are displayed on the "Pairs" page in the Tamr UI.

        Call :func:`~tamr_unify_client.dataset.resource.Dataset.refresh` from
        this dataset to regenerate pairs according to the latest binning model.

        :returns: The record pairs represented as a dataset.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """
        alias = self.api_path + "/recordPairs"
        return Dataset(self.client, None, alias)

    def pair_matching_model(self):
        """Machine learning model for pair-matching for this Mastering project.
        Learns from verified labels and predicts categorization labels for unlabeled pairs.

        Calling :func:`~tamr_unify_client.base_model.MachineLearningModel.predict`
        from this dataset will produce new (unpublished) clusters. These clusters
        are displayed on the "Clusters" page in the Tamr UI.

        :returns: The machine learning model for pair-matching.
        :rtype: :class:`~tamr_unify_client.base_model.MachineLearningModel`
        """
        alias = self.api_path + "/recordPairsWithPredictions/model"
        return MachineLearningModel(self.client, None, alias)

    def high_impact_pairs(self):
        """High-impact pairs as a dataset. Tamr labels pairs as "high-impact" if
        labeling these pairs would help it learn most quickly (i.e. "Active learning").

        High-impact pairs are displayed with a ⚡ lightning bolt icon on the
        "Pairs" page in the Tamr UI.

        Call :func:`~tamr_unify_client.dataset.resource.Dataset.refresh` from
        this dataset to produce new high-impact pairs according to the latest
        pair-matching model.

        :returns: The high-impact pairs represented as a dataset.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """
        alias = self.api_path + "/highImpactPairs"
        return Dataset(self.client, None, alias)

    def record_clusters(self):
        """Record Clusters as a dataset. Tamr clusters labeled pairs using pairs
        model. These clusters populate the cluster review page and get transient
        cluster ids, rather than published cluster ids (i.e., "Permanent Ids")

        Call :func:`~tamr_unify_client.dataset.resource.Dataset.refresh` from
        this dataset to generate clusters based on to the latest pair-matching model.

        :returns: The record clusters represented as a dataset.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """
        alias = self.api_path + "/recordClusters"
        return Dataset(self.client, None, alias)

    def published_clusters(self):
        """Published record clusters generated by Tamr's pair-matching model.

        :returns: The published clusters represented as a dataset.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """

        unified_dataset = self.unified_dataset()

        # Replace this workaround with a direct API call once API
        # is fixed. APIs that need to work are: fetching the dataset and
        # being able to call refresh on resulting dataset. Until then, we grab
        # the dataset by constructing its name from the corresponding Unified Dataset's name
        name = unified_dataset.name + "_dedup_published_clusters"
        canonical = self.client.datasets.by_name(name)
        resource_json = canonical._data
        alias = self.api_path + "/publishedClusters"
        return Dataset.from_json(self.client, resource_json, alias)

    def published_clusters_configuration(self):
        """Retrieves published clusters configuration for this project.

        :returns: The published clusters configuration
        :rtype: :class:`~tamr_unify_client.mastering.published_cluster.configuration.PublishedClustersConfiguration`
        """
        alias = self.api_path + "/publishedClustersConfiguration"
        resource_json = self.client.get(alias).successful().json()
        return PublishedClustersConfiguration.from_json(
            self.client, resource_json, alias
        )

    def published_cluster_ids(self):
        """Retrieves published cluster IDs for this project.

        :returns: The published cluster ID dataset.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """
        # Replace this workaround with a direct API call once API
        # is fixed. APIs that need to work are: fetching the dataset and
        # being able to call refresh on resulting dataset. Until then, we grab
        # the dataset by constructing its name from the corresponding Unified Dataset's name
        unified_dataset = self.unified_dataset()
        name = unified_dataset.name + "_dedup_all_persistent_ids"
        dataset = self.client.datasets.by_name(name)

        path = self.api_path + "/allPublishedClusterIds"
        return Dataset.from_json(self.client, dataset._data, path)

    def published_cluster_stats(self):
        """Retrieves published cluster stats for this project.

        :returns: The published cluster stats dataset.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """
        # Replace this workaround with a direct API call once API
        # is fixed. APIs that need to work are: fetching the dataset and
        # being able to call refresh on resulting dataset. Until then, we grab
        # the dataset by constructing its name from the corresponding Unified Dataset's name
        unified_dataset = self.unified_dataset()
        name = unified_dataset.name + "_dedup_published_cluster_stats"
        dataset = self.client.datasets.by_name(name)

        path = self.api_path + "/publishedClusterStats"
        return Dataset.from_json(self.client, dataset._data, path)

    def published_cluster_versions(self, cluster_ids):
        """Retrieves version information for the specified published clusters.
        See https://docs.tamr.com/reference#retrieve-published-clusters-given-cluster-ids.

        :param cluster_ids: The persistent IDs of the clusters to get version information for.
        :type cluster_ids: iterable[str]
        :return: A stream of the published clusters.
        :rtype: Python generator yielding :class:`~tamr_unify_client.mastering.published_cluster.resource.PublishedCluster`
        """
        path = self.api_path + "/publishedClusterVersions"
        return self._cluster_versions(PublishedCluster, cluster_ids, path)

    def record_published_cluster_versions(self, record_ids):
        """Retrieves version information for the published clusters of the given records.
        See https://docs.tamr.com/reference#retrieve-published-clusters-given-record-ids.

        :param record_ids: The Tamr IDs of the records to get cluster version information for.
        :type record_ids: iterable[str]
        :return: A stream of the relevant published clusters.
        :rtype: Python generator yielding :class:`~tamr_unify_client.mastering.published_cluster.record.RecordPublishedCluster`
        """
        path = self.api_path + "/recordPublishedClusterVersions"
        return self._cluster_versions(RecordPublishedCluster, record_ids, path)

    def _cluster_versions(self, cluster_class, ids, endpoint):
        """Retrieves version information for published clusters.

        :param cluster_class: The class to create instances of.
        :param ids: The IDs of the clusters or records to get version information for.
        :type ids: iterable[str]
        :param endpoint: The endpoint to call for versions.
        :type endpoint: str
        :return: A stream of the published clusters.
        """
        string_ids = "\n".join(json.dumps(i) for i in ids)

        with self.client.post(endpoint, data=string_ids, stream=True) as response:
            for line in response.iter_lines():
                yield cluster_class(json.loads(line))

    def estimate_pairs(self):
        """Returns pair estimate information for a mastering project

        :return: Pairs Estimate information.
        :rtype: :class:`~tamr_unify_client.mastering.estimated_pair_counts.EstimatedPairCounts`
        """
        alias = self.api_path + "/estimatedPairCounts"
        estimate_json = self.client.get(alias).successful().json()
        info = EstimatedPairCounts.from_json(self.client, estimate_json, api_path=alias)
        return info

    def record_clusters_with_data(self):
        """Project's unified dataset with associated clusters.

        :returns: The record clusters with data represented as a dataset
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """
        unified_dataset = self.unified_dataset()

        # Replace this workaround with a direct API call once API
        # is fixed. APIs that need to work are: fetching the dataset and
        # being able to call refresh on resulting dataset. Until then, we grab
        # the dataset by constructing its name from the corresponding Unified Dataset's name
        name = unified_dataset.name + "_dedup_clusters_with_data"
        dataset = self.client.datasets.by_name(name)
        dataset.api_path = self.api_path + "/recordClustersWithData"
        return dataset

    def published_clusters_with_data(self):
        """Project's unified dataset with associated clusters.

        :returns: The published clusters with data represented as a dataset
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """

        unified_dataset = self.unified_dataset()
        name = unified_dataset.name + "_dedup_published_clusters_with_data"
        dataset = self.client.datasets.by_name(name)
        dataset.api_path = self.api_path + "/publishedClustersWithData"
        return dataset

    def binning_model(self):
        """
        Binning model for this project.

        :return: Binning model for this project.
        :rtype: :class:`~tamr_unify_client.mastering.binning_model.BinningModel`
        """
        alias = self.api_path + "/binningModel"

        # Cannot get this resource and so we hard code
        resource_json = {"relativeId": alias}
        return BinningModel.from_json(self.client, resource_json, alias)

    # super.__repr__ is sufficient



from dataclasses import dataclass
from typing import Optional, Union

from tamr_client._types.attribute import Attribute
from tamr_client._types.url import URL


@dataclass(frozen=True)
class CategorizationProject:
    """A Tamr Categorization project

    See https://docs.tamr.com/reference/the-project-object

    Args:
        url
        name
        description
    """

    url: URL
    name: str
    description: Optional[str] = None


@dataclass(frozen=True)
class MasteringProject:
    """A Tamr Mastering project

    See https://docs.tamr.com/reference/the-project-object

    Args:
        url
        name
        description
    """

    url: URL
    name: str
    description: Optional[str] = None


@dataclass(frozen=True)
class SchemaMappingProject:
    """A Tamr Schema Mapping project

    See https://docs.tamr.com/reference/the-project-object

    Args:
        url
        name
        description
    """

    url: URL
    name: str
    description: Optional[str] = None


@dataclass(frozen=True)
class GoldenRecordsProject:
    """A Tamr Golden Records project

    See https://docs.tamr.com/reference/the-project-object

    Args:
        url
        name
        description
    """

    url: URL
    name: str
    description: Optional[str] = None


@dataclass(frozen=True)
class UnknownProject:
    """A Tamr project of an unrecognized type

    See https://docs.tamr.com/reference/the-project-object

    Args:
        url
        name
        description
    """

    url: URL
    name: str
    description: Optional[str] = None


Project = Union[
    CategorizationProject,
    MasteringProject,
    SchemaMappingProject,
    GoldenRecordsProject,
    UnknownProject,
]


@dataclass(frozen=True)
class AttributeMapping:
    """A Tamr Attribute Mapping.

    See https://docs.tamr.com/new/reference/retrieve-projects-mappings

    Args:
        url
        input_attribute
        unified_attribute
    """

    url: URL
    input_attribute: Attribute
    unified_attribute: Attribute



class RecordPublishedClusterVersion:
    """A version of a published cluster in a mastering project.

    This is not a `BaseResource` because it does not have its own API endpoint.

    :param data: The JSON entity representing this version.
    """

    def __init__(self, data):
        self._data = data

    @property
    def version(self):
        """:type: str"""
        return self._data.get("version")

    @property
    def timestamp(self):
        """:type: str"""
        return self._data.get("timestamp")

    @property
    def cluster_id(self):
        """:type: str"""
        return self._data.get("clusterId")

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"version={self.version!r}, "
            f"timestamp={self.timestamp!r}, "
            f"name={self.cluster_id!r})"
        )



import logging
from typing import Optional
from urllib.parse import urlparse

import requests
import requests.auth
import requests.exceptions

from tamr_unify_client.auth.token import TokenAuth
from tamr_unify_client.dataset.collection import DatasetCollection
from tamr_unify_client.project.collection import ProjectCollection
import tamr_unify_client.response as response

logger = logging.getLogger(__name__)

response._monkey_patch()


class Client:
    """Python Client for Tamr API.

    Each client is specific to a specific origin (protocol, host, port).

    Args:
        auth: Tamr-compatible Authentication provider.

            **Recommended**: use one of the classes described in :ref:`authentication`
        host: Host address of remote Tamr instance (e.g. ``'10.0.10.0'``)
        protocol: Either ``'http'`` or ``'https'``
        port: Tamr instance main port
        base_path: Base API path. Requests made by this client will be relative to this path.
        session: Session to use for API calls. If none is provided, will use a new :class:`requests.Session`.

    Example:
        >>> from tamr_unify_client import Client
        >>> from tamr_unify_client.auth import UsernamePasswordAuth
        >>> auth = UsernamePasswordAuth('my username', 'my password')
        >>> tamr_local = Client(auth) # on http://localhost:9100
        >>> tamr_remote = Client(auth, protocol='https', host='10.0.10.0') # on https://10.0.10.0:9100
        >>> tamr_remote = Client(auth, protocol='https', host='10.0.10.0', port=None) # on https://10.0.10.0
    """

    def __init__(
        self,
        auth: requests.auth.AuthBase,
        host: str = "localhost",
        protocol: str = "http",
        port: Optional[int] = 9100,
        base_path: str = "/api/versioned/v1/",
        session: Optional[requests.Session] = None,
        store_auth_cookie: bool = False,
    ):
        self.auth = auth
        self.host = host
        self.protocol = protocol
        self.port = port
        self.base_path = base_path
        self.session = session or requests.Session()
        self.session.auth = auth
        if store_auth_cookie:
            self.set_auth_cookie()

        self._projects = ProjectCollection(self)
        self._datasets = DatasetCollection(self)

        if not self.base_path.startswith("/"):
            self.base_path = "/" + self.base_path

        if not self.base_path.endswith("/"):
            self.base_path = self.base_path + "/"

    @property
    def origin(self) -> str:
        """HTTP origin i.e. :code:`<protocol>://<host>[:<port>]`.

        For additional information, see `MDN web docs <https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Origin>`_ .
        """
        if self.port is None:
            return f"{self.protocol}://{self.host}"
        else:
            return f"{self.protocol}://{self.host}:{self.port}"

    def request(self, method: str, endpoint: str, **kwargs) -> requests.Response:
        """Sends a request to Tamr.

        The URL for the request will be ``<origin>/<base_path>/<endpoint>``.
        The request is authenticated via :attr:`Client.auth`.

        Args:
            method: The HTTP method to use (e.g. `'GET'` or `'POST'`)
            endpoint: API endpoint to call (relative to the Base API path for this client).

        Returns:
            HTTP response from the Tamr server
        """
        if urlparse(endpoint).scheme != "" and urlparse(endpoint).netloc != "":
            url = endpoint
        elif endpoint.startswith("/"):
            url = self.origin + endpoint
        else:
            url = self.origin + self.base_path + endpoint

        # Attempt request with auth cookie
        response = self.session.request(method, url, **kwargs)
        if response.status_code == 401 and "credentials" in response.text.lower():
            first_response = response
            self.set_auth_cookie()
            response = self.session.request(method, url, **kwargs)
            if response.status_code == 401 and "credentials" in response.text.lower():
                # Login credentials are bad, return original response
                response = first_response

        logger.info(
            f"{response.request.method} {response.url} : {response.status_code}"
        )
        return response

    def get(self, endpoint, **kwargs):
        """Calls :func:`~tamr_unify_client.Client.request` with the ``"GET"`` method.
        """
        return self.request("GET", endpoint, **kwargs)

    def post(self, endpoint, **kwargs):
        """Calls :func:`~tamr_unify_client.Client.request` with the ``"POST"`` method.
        """
        return self.request("POST", endpoint, **kwargs)

    def put(self, endpoint, **kwargs):
        """Calls :func:`~tamr_unify_client.Client.request` with the ``"PUT"`` method.
        """
        return self.request("PUT", endpoint, **kwargs)

    def delete(self, endpoint, **kwargs):
        """Calls :func:`~tamr_unify_client.Client.request` with the ``"DELETE"`` method.
        """
        return self.request("DELETE", endpoint, **kwargs)

    def set_auth_cookie(self):
        """Fetch and store an auth token for the given client configuration"""
        # Fetch auth token and store as cookie
        if isinstance(self.auth, TokenAuth):
            raise TypeError("Auth cookie not supported for TokenAuth authentication")
        r = self.post(
            "./instance:login",
            json={"username": self.auth.username, "password": self.auth.password},
        )
        # If login request fails for any reason do not set cookie so following responses use
        # header credentials for authentication
        if not r.ok:
            return
        auth_token = r.json()["token"]
        self.session.cookies.set("authToken", auth_token)  # TODO: Set domain
        # Clear session auth.  It is still accessible as self.auth if needed
        self.session.auth = None

    @property
    def projects(self) -> ProjectCollection:
        """Collection of all projects on this Tamr instance.

        Returns:
            Collection of all projects.
        """
        return self._projects

    @property
    def datasets(self) -> DatasetCollection:
        """Collection of all datasets on this Tamr instance.

        Returns:
            Collection of all datasets.
        """
        return self._datasets

    def __repr__(self):
        # Show only the type `auth` to mitigate any security concerns.
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"host={self.host!r}, "
            f"protocol={self.protocol!r}, "
            f"port={self.port!r}, "
            f"base_path={self.base_path!r}, "
            f"auth={type(self.auth).__name__})"
        )



from dataclasses import dataclass
from typing import Optional, Tuple, Union

from tamr_client._types.url import URL


@dataclass(frozen=True)
class Dataset:
    """A Tamr dataset

    See https://docs.tamr.com/reference/dataset-models

    Args:
        url: The canonical dataset-based URL for this dataset e.g. `/datasets/1`
        name
        key_attribute_names
        description
    """

    url: URL
    name: str
    key_attribute_names: Tuple[str, ...]
    description: Optional[str] = None


@dataclass(frozen=True)
class UnifiedDataset:
    """A Tamr unified dataset

    See https://docs.tamr.com/reference/dataset-models

    Args:
        url: The project-based alias for this dataset e.g. `/projects/1/unifiedDataset`
        name
        key_attribute_names
        description
    """

    url: URL
    name: str
    key_attribute_names: Tuple[str, ...]
    description: Optional[str] = None


AnyDataset = Union[Dataset, UnifiedDataset]



from typing import Optional

import requests

from tamr_client._types.auth import UsernamePasswordAuth


class Session(requests.Session):
    def __init__(self):
        super(self.__class__, self).__init__()
        self._stored_auth: Optional[UsernamePasswordAuth] = None

    def request(self, *args, **kwargs):
        # signature of `requests` requires not naming positional args
        response = super(self.__class__, self).request(*args, **kwargs)
        if response.status_code == 401 and "credentials" in response.text.lower():
            first_response = response
            self._set_auth_cookie(args[1])
            response = super(self.__class__, self).request(*args, **kwargs)
            if response.status_code == 401 and "credentials" in response.text.lower():
                # Login credentials are bad, return original response
                response = first_response
        return response

    def _set_auth_cookie(self, parent_url: str):
        """Fetch and store an auth token for the given client configuration"""
        # No-op if _stored_auth is None
        if self._stored_auth is None:
            return

        # Fetch auth token and store as cookie
        socket_address = parent_url.split("/api/")[0]
        r = self.post(
            socket_address + "/api/versioned/v1/instance:login",
            json={
                "username": self._stored_auth.username,
                "password": self._stored_auth.password,
            },
        )
        if r.status_code == 200:
            auth_token = r.json()["token"]
            self.cookies.set("authToken", auth_token)  # TODO: Set domain for security
            # Clear session auth if cookie is retrieved
            self.auth = None
        else:
            # Set session auth from client auth in case it has been cleared
            # Allow the following call to pass credentials in header
            self.auth = self._stored_auth



from tamr_unify_client.base_resource import BaseResource
from tamr_unify_client.categorization.category.collection import CategoryCollection


class Taxonomy(BaseResource):
    """A project's taxonomy"""

    @classmethod
    def from_json(cls, client, data, api_path):
        return super().from_data(client, data, api_path)

    @property
    def name(self):
        """:type: str"""
        return self._data.get("name")

    def categories(self):
        """Retrieves the categories of this taxonomy.

        :returns: A collection of the taxonomy categories.
        :rtype: :class:`~tamr_unify_client.categorization.category.collection.CategoryCollection`
        """
        alias = self.api_path + "/categories"
        return CategoryCollection(self.client, alias)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"relative_id={self.relative_id!r}, "
            f"name={self.name!r})"
        )



import json
import logging
from typing import Iterator

import requests

from tamr_client._types import JsonDict

logger = logging.getLogger(__name__)


def successful(response: requests.Response) -> requests.Response:
    """Ensure response does not contain an HTTP error.

    Delegates to :func:`requests.Response.raise_for_status`

    Returns:
        The response being checked.

    Raises:
        requests.exceptions.HTTPError: If an HTTP error is encountered.
    """
    try:
        response.raise_for_status()
    except requests.HTTPError as e:
        r = e.response
        logger.error(
            f"Encountered HTTP error code {r.status_code}. Response body: {r.text}"
        )
        raise e
    return response


def ndjson(response: requests.Response, **kwargs) -> Iterator[JsonDict]:
    """Stream newline-delimited JSON from the response body

    Analog to :func:`requests.Response.json` but for ``.ndjson``-formatted body.

    **Recommended**: For memory efficiency, use ``stream=True`` when sending the request corresponding to this response.

    Args:
        response: Response whose body should be streamed as newline-delimited JSON.
        **kwargs: Keyword arguments passed to underlying :func:`requests.Response.iter_lines` call.

    Returns
        Each line of the response body, parsed as JSON

    Example:
        >>> import tamr_client as tc
        >>> s = tc.session.from_auth(...)
        >>> r = s.get(..., stream=True)
        >>> for data in tc.response.ndjson(r):
        ...     assert data['my key'] == 'my_value'

    """
    for line in response.iter_lines(**kwargs):
        yield json.loads(line)



import base64

import requests


def _basic_auth_str(username, password):
    auth = f"{username}:{password}"
    encoded = base64.b64encode(auth.encode("latin1"))
    return "BasicCreds " + requests.utils.to_native_string(encoded.strip())


class UsernamePasswordAuth(requests.auth.HTTPBasicAuth):
    """Provides username/password authentication for Tamr.

    Sets the `Authorization` HTTP header with Tamr's custom `BasicCreds` format.

    Args:
        username:
        password:

    Example:
        >>> import tamr_client as tc
        >>> auth = tc.UsernamePasswordAuth('my username', 'my password')
        >>> s = tc.Session(auth)
    """

    def __init__(self, username: str, password: str):
        super().__init__(username, password)

    def __call__(self, r):
        r.headers["Authorization"] = _basic_auth_str(self.username, self.password)
        return r

    def __repr__(self):
        # intentionally leave out password (potentially sensitive)
        return (
            f"{type(self).__qualname__}("
            f"username={repr(self.username)}"
            f"password=<redacted>)"
        )



"""
See https://docs.tamr.com/reference/record
"The recommended approach for modifying records is to use the :func:`~tamr_client.record.upsert` and
:func:`~tamr_client.record.delete` functions for all use cases they can handle. For more advanced use cases, the
underlying :func:`~tamr_client.record._update` function can be used directly."
"""
import json
from typing import cast, Dict, IO, Iterable, Iterator, Optional

from tamr_client import primary_key
from tamr_client import response
from tamr_client._types import AnyDataset, Dataset, JsonDict, Session


def _update(session: Session, dataset: Dataset, updates: Iterable[Dict]) -> JsonDict:
    """Send a batch of record creations/updates/deletions to this dataset.
    You probably want to use :func:`~tamr_client.record.upsert`
    or :func:`~tamr_client.record.delete` instead.

    Args:
        dataset: Dataset containing records to be updated
        updates: Each update should be formatted as specified in the `Public Docs for Dataset updates <https://docs.tamr.com/reference#modify-a-datasets-records>`_.

    Returns:
        JSON response body from server

    Raises:
        requests.HTTPError: If an HTTP error is encountered
    """
    stringified_updates = (json.dumps(update).encode("utf-8") for update in updates)
    # `requests` accepts a generator for `data` param, but stubs for `requests` in https://github.com/python/typeshed expects this to be a file-like object
    io_updates = cast(IO, stringified_updates)
    r = session.post(
        str(dataset.url) + ":updateRecords",
        headers={"Content-Encoding": "utf-8"},
        data=io_updates,
    )
    return response.successful(r).json()


def upsert(
    session: Session,
    dataset: Dataset,
    records: Iterable[Dict],
    *,
    primary_key_name: Optional[str] = None,
) -> JsonDict:
    """Create or update the specified records.

    Args:
        dataset: Dataset to receive record updates
        records: The records to update, as dictionaries
        primary_key_name: The primary key for these records, which must be a key in each record dictionary.
            By default the key_attribute_name of dataset

    Returns:
        JSON response body from server

    Raises:
        requests.HTTPError: If an HTTP error is encountered
        primary_key.NotFound: If primary_key_name does not match dataset primary key
        primary_key.NotFound: If primary_key_name not in a record dictionary
    """
    if primary_key_name is None:
        primary_key_name = dataset.key_attribute_names[0]

    if primary_key_name not in dataset.key_attribute_names:
        raise primary_key.NotFound(
            f"Primary key: {primary_key_name} is not in dataset key attribute names: {dataset.key_attribute_names}"
        )
    updates = (
        _create_command(record, primary_key_name=primary_key_name) for record in records
    )
    return _update(session, dataset, updates)


def delete(
    session: Session,
    dataset: Dataset,
    records: Iterable[Dict],
    *,
    primary_key_name: Optional[str] = None,
) -> JsonDict:
    """Deletes the specified records, based on primary key values.  Does not check that other attribute values match.

    Args:
        dataset: Dataset from which to delete records
        records: The records to update, as dictionaries
        primary_key_name: The primary key for these records, which must be a key in each record dictionary.
            By default the key_attribute_name of dataset

    Returns:
        JSON response body from server

    Raises:
        requests.HTTPError: If an HTTP error is encountered
        primary_key.NotFound: If primary_key_name does not match dataset primary key
        primary_key.NotFound: If primary_key_name not in a record dictionary
    """
    if primary_key_name is None:
        primary_key_name = dataset.key_attribute_names[0]

    if primary_key_name not in dataset.key_attribute_names:
        raise primary_key.NotFound(
            f"Primary key: {primary_key_name} is not in dataset key attribute names: {dataset.key_attribute_names}"
        )
    updates = (
        _delete_command(record, primary_key_name=primary_key_name) for record in records
    )
    return _update(session, dataset, updates)


def _create_command(record: Dict, *, primary_key_name: str) -> Dict:
    """Generates the CREATE command formatted as specified in the `Public Docs for Dataset updates
    <https://docs.tamr.com/reference#modify-a-datasets-records>`_.

    Args:
        record: The record to create, as a dictionary
        primary_key_name: The primary key for this record, which must be a key in the dictionary

    Returns:
        The CREATE command in the proper format
    """
    return {"action": "CREATE", "recordId": record[primary_key_name], "record": record}


def _delete_command(record: Dict, *, primary_key_name: str) -> Dict:
    """Generates the DELETE command formatted as specified in the `Public Docs for Dataset updates
    <https://docs.tamr.com/reference#modify-a-datasets-records>`_.

    Args:
        record: The record to delete, as a dictionary
        primary_key_name: The primary key for this record, which must be a key in the dictionary

    Returns:
        The DELETE command in the proper format
    """
    return {"action": "DELETE", "recordId": record[primary_key_name]}


def stream(session: Session, dataset: AnyDataset) -> Iterator[JsonDict]:
    """Stream the records in this dataset as Python dictionaries.

    Args:
        dataset: Dataset from which to stream records

    Returns:
        Python generator yielding records
    """
    with session.get(str(dataset.url) + "/records", stream=True) as r:
        yield from response.ndjson(r)


def delete_all(session: Session, dataset: AnyDataset):
    """Delete all records in this dataset

    Args:
        dataset: Dataset from which to delete records
    """
    r = session.delete(str(dataset.url) + "/records")
    response.successful(r)



from dataclasses import dataclass

from tamr_client._types.instance import Instance


@dataclass(frozen=True)
class URL:
    path: str
    instance: Instance = Instance()
    base_path: str = "api/versioned/v1"

    def __str__(self):
        from tamr_client import instance

        origin = instance.origin(self.instance)
        return f"{origin}/{self.base_path}/{self.path}"



from abc import abstractmethod
from collections.abc import Iterable


class BaseCollection(Iterable):
    """Base class for client-side collections.

    :param client: Delegate underlying API calls to this client.
    :type client: :class:`~tamr_unify_client.Client`
    :param api_path: API path for this collection. E.g. ``"projects/1/inputDatasets"``.
    :type api_path: str
    """

    def __init__(self, client, api_path):
        self.client = client
        self.api_path = api_path

    @abstractmethod
    def by_resource_id(self, canonical_path, resource_id):
        """Retrieve an item in this collection by resource ID.

        Subclasses should override this method and pass in the specific
        ``canonical_path`` for that collection.

        :param canonical_path: The canonical (i.e. unaliased) API path for this collection.
        :type canonical_path: str
        :param resource_id: The resource ID. E.g. "1"
        :type resource_id: str
        :returns: The specified item.
        :rtype: The ``resource_class``  for this collection. See :func:`~tamr_unify_client.base_collection.BaseCollection.by_relative_id`.
        """
        relative_id = canonical_path + "/" + resource_id
        return self.by_relative_id(relative_id)

    @abstractmethod
    def by_relative_id(self, resource_class, relative_id):
        """Retrieve an item in this collection by relative ID.

        Subclasses should override this method and pass in the specific
        ``resource_class`` for that collection.

        :param resource_class: Resource class corresponding to items in this collection.
        :type resource_class: Python class
        :param relative_id: The relative ID. E.g. "projects/1"
        :type relative_id: str
        :returns: The specified item.
        :rtype: ``resource_class``
        """
        resource_json = self.client.get(relative_id).successful().json()
        return resource_class.from_json(
            self.client, resource_json, api_path=relative_id
        )

    @abstractmethod
    def stream(self, resource_class):
        """Stream items in this collection.

        Subclasses should override this method and pass in the specific
        ``resource_class`` for that collection.

        :param resource_class: Resource class corresponding to items in this collection.
        :type resource_class: Python class
        :returns: Generator that yields each item.
        :rtype: Python generator of ``resource_class``
        """
        resources = self.client.get(self.api_path).successful().json()
        for resource_json in resources:
            yield resource_class.from_json(self.client, resource_json)

    def __iter__(self):
        return self.stream()

    @abstractmethod
    def by_external_id(self, resource_class, external_id):
        """Retrieve an item in this collection by external ID.

        Subclasses should override this method and pass in the specific
        ``resource_class`` for that collection.

        :param resource_class: Resource class corresponding to items in this collection.
        :type resource_class: Python class
        :param external_id: The external ID.
        :type external_id: str
        :returns: The specified item, if found.
        :rtype: ``resource_class``
        :raises KeyError: If no resource with the specified external_id is found
        :raises LookupError: If multiple resources with the specified external_id are found
        """
        params = {"filter": "externalId==" + external_id}
        resources = self.client.get(self.api_path, params=params).successful().json()
        items = [
            resource_class.from_json(self.client, resource_json)
            for resource_json in resources
        ]

        if len(items) == 0:
            raise KeyError(f'No item found with external ID "{external_id}"')
        elif len(items) > 1:
            raise LookupError(
                f'More than one item found with external ID "{external_id}"'
            )

        return items[0]

    def delete_by_resource_id(self, resource_id):
        """Deletes a resource from this collection by resource ID.

        :param resource_id: The resource ID of the resource that will be deleted.
        :type resource_id: str
        :return: HTTP response from the server.
        :rtype: :class:`requests.Response`
        """
        path = f"{self.api_path}/{resource_id}"
        response = self.client.delete(path).successful()
        return response

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"api_path={self.api_path!r})"
        )



from dataclasses import dataclass

from tamr_client._types.url import URL


@dataclass(frozen=True)
class Backup:
    """A Tamr backup

    See https://docs.tamr.com/new/docs/configuration-backup-and-restore

    Args:
        url
        path
        state
        error_message
    """

    url: URL
    path: str
    state: str
    error_message: str



import requests

from tamr_client import dataset, response
from tamr_client._types import (
    InputTransformation,
    Instance,
    JsonDict,
    Project,
    Session,
    Transformations,
)


def _input_transformation_from_json(
    session: Session, instance: Instance, data: JsonDict
) -> InputTransformation:
    """Make input transformation from JSON data (deserialize)

    Args:
        instance: Tamr instance containing this transformation
        data: Input scoped transformation JSON data from Tamr server
    """
    dataset_resource_ids = [d["datasetId"].split("/")[-1] for d in data["datasets"]]
    datasets = [
        dataset.by_resource_id(session, instance, d_id) for d_id in dataset_resource_ids
    ]
    return InputTransformation(transformation=data["transformation"], datasets=datasets)


def _from_json(session: Session, instance: Instance, data: JsonDict) -> Transformations:
    """Make transformations from JSON data (deserialize)

    Args:
        instance: Tamr instance containing this transformation
        data: Transformation JSON data from Tamr server
    """
    return Transformations(
        unified_scope=data["unified"],
        input_scope=[
            _input_transformation_from_json(session, instance, tx)
            for tx in data["parameterized"]
        ],
    )


def _input_transformation_to_json(tx: InputTransformation) -> JsonDict:
    """Convert input transformations to JSON data (serialize)

    Args:
        tx: Input transformation to convert
    """
    # datasetId omitted, only one of "datasetId" or "relativeDatasetId" is required
    dataset_json = [
        {"name": d.name, "relativeDatasetId": d.url.path} for d in tx.datasets
    ]

    return {"datasets": dataset_json, "transformation": tx.transformation}


def _to_json(tx: Transformations) -> JsonDict:
    """Convert transformations to JSON data (serialize)

    Args:
        tx: Transformations to convert
    """
    return {
        "parameterized": [_input_transformation_to_json(t) for t in tx.input_scope],
        "unified": tx.unified_scope,
    }


def get_all(session: Session, project: Project) -> Transformations:
    """Get the transformations of a Project

    Args:
        project: Project containing transformations

    Raises:
        requests.HTTPError: If any HTTP error is encountered.

    Example:
        >>> import tamr_client as tc
        >>> session = tc.session.from_auth('username', 'password')
        >>> instance = tc.instance.Instance(host="localhost", port=9100)
        >>> project1 = tc.project.by_resource_id(session, instance, id='1')
        >>> print(tc.transformations.get_all(session, project1))
    """
    r = session.get(f"{project.url}/transformations")
    response.successful(r)
    return _from_json(session, project.url.instance, r.json())


def replace_all(
    session: Session, project: Project, tx: Transformations
) -> requests.Response:
    """Replaces the transformations of a Project

    Args:
        project: Project to place transformations within
        tx: Transformations to put into project

    Raises:
        requests.HTTPError: If any HTTP error is encountered.

    Example:
        >>> import tamr_client as tc
        >>> session = tc.session.from_auth('username', 'password')
        >>> instance = tc.instance.Instance(host="localhost", port=9100)
        >>> project1 = tc.project.by_resource_id(session, instance, id='1')
        >>> dataset3 = tc.dataset.by_resource_id(session, instance, id='3')
        >>> new_input_tx = tc.InputTransformation("SELECT *, upper(name) as name;", [dataset3])
        >>> all_tx = tc.Transformations(
        ... input_scope=[new_input_tx],
        ... unified_scope=["SELECT *, 1 as one;"]
        ... )
        >>> tc.transformations.replace_all(session, project1, all_tx)
    """
    body = _to_json(tx)
    r = session.put(f"{project.url}/transformations", json=body)

    return response.successful(r)



"""
Tamr - Schema Mapping
See https://docs.tamr.com/new/docs/overall-workflow-schema

The terminology used here is consistent with Tamr UI terminology

Asynchronous versions of each function can be found with the suffix `_async` and may be of
interest to power users
"""
from tamr_client import operation
from tamr_client._types import Operation, SchemaMappingProject, Session
from tamr_client.dataset import unified


def update_unified_dataset(
    session: Session, project: SchemaMappingProject
) -> Operation:
    """Apply changes to the unified dataset and wait for the operation to complete

    Args:
        project: Tamr Schema Mapping project
    """
    unified_dataset = unified.from_project(session, project)
    op = unified._apply_changes_async(session, unified_dataset)
    return operation.wait(session, op)



import json

from tamr_unify_client.base_collection import BaseCollection
from tamr_unify_client.categorization.category.resource import Category


class CategoryCollection(BaseCollection):
    """Collection of :class:`~tamr_unify_client.categorization.category.resource.Category` s.

    :param client: Client for API call delegation.
    :type client: :class:`~tamr_unify_client.Client`
    :param api_path: API path used to access this collection.
        E.g. ``"projects/1/taxonomy/categories"``.
    :type api_path: str
    """

    def __init__(self, client, api_path):
        super().__init__(client, api_path)

    def by_resource_id(self, resource_id):
        """Retrieve a  category by resource ID.

        :param resource_id: The resource ID. E.g. ``"1"``
        :type resource_id: str
        :returns: The specified category.
        :rtype: :class:`~tamr_unify_client.categorization.category.resource.Category`
        """
        return super().by_resource_id(self.api_path, resource_id)

    def by_relative_id(self, relative_id):
        """Retrieve a category by relative ID.

        :param relative_id: The relative ID. E.g. ``"projects/1/categories/1"``
        :type relative_id: str
        :returns: The specified category.
        :rtype: :class:`~tamr_unify_client.categorization.category.resource.Category`
        """
        return super().by_relative_id(Category, relative_id)

    def by_external_id(self, external_id):
        """Retrieve an attribute by external ID.

        Since categories do not have external IDs, this method is not supported and will
        raise a :class:`NotImplementedError` .

        :param external_id: The external ID.
        :type external_id: str
        :returns: The specified category, if found.
        :rtype: :class:`~tamr_unify_client.categorization.category.resource.Category`
        :raises KeyError: If no category with the specified external_id is found
        :raises LookupError: If multiple categories with the specified external_id are found
        """
        raise NotImplementedError("Categories do not have external_id")

    def stream(self):
        """Stream categories in this collection. Implicitly called when iterating
        over this collection.

        :returns: Stream of categories.
        :rtype: Python generator yielding :class:`~tamr_unify_client.categorization.category.resource.Category`

        Usage:
            >>> for category in collection.stream(): # explicit
            >>>     do_stuff(category)
            >>> for category in collection: # implicit
            >>>     do_stuff(category)
        """
        return super().stream(Category)

    def create(self, creation_spec):
        """ Creates a new category.

        :param creation_spec: Category creation specification, formatted as specified in the
            `Public Docs for Creating a Category <https://docs.tamr.com/reference#create-a-category>`_.
        :type creation_spec: dict
        :return: The newly created category.
        :rtype: :class:`~tamr_unify_client.categorization.category.resource.Category`
        """
        resource_json = (
            self.client.post(self.api_path, json=creation_spec).successful().json()
        )
        return Category.from_json(self.client, resource_json)

    def bulk_create(self, creation_specs):
        """Creates new categories in bulk.

        :param creation_specs: A collection of creation specifications, as detailed for create.
        :type creation_specs: iterable[dict]
        :returns: JSON response from the server
        :rtype: :py:class:`dict`
        """
        body = "\n".join([json.dumps(s) for s in creation_specs]).encode("utf-8")
        return (
            self.client.post(
                self.api_path + ":bulk",
                headers={"Content-Encoding": "utf-8"},
                data=body,
            )
            .successful()
            .json()
        )

    # super.__repr__ is sufficient



from tamr_client._types import Session
from tamr_client._types.auth import UsernamePasswordAuth


def from_auth(auth: UsernamePasswordAuth) -> Session:
    """Create a new authenticated session

    Args:
        auth: Authentication
    """
    s = Session()
    s._stored_auth = auth
    return s



from dataclasses import dataclass
from typing import Dict, Optional

from tamr_client._types.url import URL


@dataclass(frozen=True)
class Operation:
    """A Tamr operation

    See https://docs.tamr.com/new/reference/the-operation-object

    Args:
        url
        type
        status
        description
    """

    url: URL
    type: str
    status: Optional[Dict[str, str]] = None
    description: Optional[str] = None



from copy import deepcopy

from tamr_unify_client.base_resource import BaseResource


class PublishedClustersConfiguration(BaseResource):
    """
    The configuration of published clusters in a project.

    See https://docs.tamr.com/reference#the-published-clusters-configuration-object
    """

    @classmethod
    def from_json(cls, client, data, api_path):
        return super().from_data(client, data, api_path)

    @property
    def relative_id(self):
        """:type: str"""
        # api_path is alias when it exists, and relative_id when it does not.
        # this distinction is useful for things like refreshing a unified dataset,
        # where using the relative_id would hit
        #   /datasets/{id}:refresh
        # rather than
        #   /projects/{id}/unifiedDataset:refresh.
        # Since cluster configurations don't currently have that kind of aliasing,
        # using api_path is always correct.
        # If configurations ever get aliased, this will need to be updated.
        # This is confusing; there's an RFC for suggestions to improve this
        # #64 https://github.com/Datatamer/unify-client-python/issues/64
        # "Conflation between 'api_path', 'relative_id' / 'relativeId', and
        # BaseResource ctor 'alias'"
        return self.api_path

    @property
    def versions_time_to_live(self):
        """:type: str"""
        return self._data.get("versionsTimeToLive")

    def spec(self):
        """Returns a spec representation of this published cluster configuration.

        :return: The published cluster configuration spec.
        :rtype: :class`~tamr_unify_client.mastering.published_cluster.configuration.PublishedClustersConfigurationSpec`
        """
        return PublishedClustersConfigurationSpec.of(self)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"relative_id={self.relative_id!r}, "
            f"versions_time_to_live={self.versions_time_to_live!r})"
        )


class PublishedClustersConfigurationSpec:
    """A representation of the server view of published clusters configuration."""

    def __init__(self, client, data, api_path):
        self.client = client
        self._data = data
        self.api_path = api_path

    @staticmethod
    def of(resource):
        """Creates an published cluster configuration spec from published cluster configuration.

        :param resource: The existing published cluster configuration.
        :type resource: :class:`~tamr_unify_client.mastering.published_cluster.configuration.PublishedClustersConfiguration`
        :return: The corresponding published cluster configuration spec.
        :rtype: :class:`~tamr_unify_client.mastering.published_cluster.configuration.PublishedClustersConfigurationSpec`
        """
        return PublishedClustersConfigurationSpec(
            resource.client, deepcopy(resource._data), resource.api_path
        )

    def from_data(self, data):
        """Creates a spec with new data.

        :param data: The data for the new spec.
        :type data: dict
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.mastering.published_cluster.configuration.PublishedClustersConfigurationSpec`
        """
        return PublishedClustersConfigurationSpec(self.client, data, self.api_path)

    def to_dict(self):
        """Returns a version of this spec that conforms to the API representation.

        :returns: The spec's dict.
        :rtype: dict
        """
        return deepcopy(self._data)

    def with_versions_time_to_live(self, new_versions_time_to_live):
        """Creates a new spec with the same properties, updating versions time to live.

        :param new_versions_time_to_live: The new versions time to live.
        :type new_versions_time_to_live: str
        :return: A new spec.
        :rtype: :class:`~tamr_unify_client.mastering.published_cluster.configuration.PublishedClustersConfigurationSpec`
        """
        return self.from_data(
            {**self._data, "versionsTimeToLive": new_versions_time_to_live}
        )

    def put(self):
        """Commits these changes by updating the configuration in Tamr.

        :return: The updated configuration.
        :rtype: :class:`~tamr_unify_client.mastering.published_cluster.configuration.PublishedClustersConfiguration`
        """
        updated_json = (
            self.client.put(self.api_path, json=self._data).successful().json()
        )
        return PublishedClustersConfiguration.from_json(
            self.client, updated_json, self.api_path
        )

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"dict={self._data})"
        )



from typing import List, Optional, Tuple, Union

from tamr_client import response
from tamr_client._types import (
    Attribute,
    Instance,
    JsonDict,
    Project,
    Session,
    UnknownProject,
    URL,
)
from tamr_client.attribute import _get_all_from_parent
from tamr_client.categorization import project as categorization_project
from tamr_client.exception import TamrClientException
from tamr_client.golden_records import project as golden_records_project
from tamr_client.mastering import project as mastering_project
from tamr_client.schema_mapping import project as schema_mapping_project


class NotFound(TamrClientException):
    """Raised when referencing (e.g. updating or deleting) a project
    that does not exist on the server."""

    pass


class Ambiguous(TamrClientException):
    """Raised when referencing a project by name that matches multiple possible targets."""

    pass


class AlreadyExists(TamrClientException):
    """Raised when a project with these specifications already exists."""

    pass


def by_resource_id(session: Session, instance: Instance, id: str) -> Project:
    """Get project by resource ID.
    Fetches project from Tamr server.

    Args:
        instance: Tamr instance containing this dataset
        id: Project ID

    Raises:
        project.NotFound: If no project could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    url = URL(instance=instance, path=f"projects/{id}")
    return _by_url(session, url)


def by_name(session: Session, instance: Instance, name: str) -> Project:
    """Get project by name
    Fetches project from Tamr server.

    Args:
        instance: Tamr instance containing this project
        name: Project name

    Raises:
        project.NotFound: If no project could be found with that name.
        project.Ambiguous: If multiple targets match project name.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    r = session.get(
        url=str(URL(instance=instance, path="projects")),
        params={"filter": f"name=={name}"},
    )

    # Check that exactly one project is returned
    matches = r.json()
    if len(matches) == 0:
        raise NotFound(str(r.url))
    if len(matches) > 1:
        raise Ambiguous(str(r.url))

    # Make Project from response
    url = URL(instance=instance, path=matches[0]["relativeId"])
    return _from_json(url=url, data=matches[0])


def _by_url(session: Session, url: URL) -> Project:
    """Get project by URL.
    Fetches project from Tamr server.

    Args:
        url: Project URL

    Raises:
        project.NotFound: If no project could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    r = session.get(str(url))
    if r.status_code == 404:
        raise NotFound(str(url))
    data = response.successful(r).json()
    return _from_json(url, data)


def _from_json(url: URL, data: JsonDict) -> Project:
    """Make project from JSON data (deserialize)
    Args:
        url: Project URL
        data: Project JSON data from Tamr server
    """
    proj_type = data["type"]
    if proj_type == "DEDUP":
        return mastering_project._from_json(url, data)
    elif proj_type == "CATEGORIZATION":
        return categorization_project._from_json(url, data)
    elif proj_type == "SCHEMA_MAPPING_RECOMMENDATIONS":
        return schema_mapping_project._from_json(url, data)
    elif proj_type == "GOLDEN_RECORDS":
        return golden_records_project._from_json(url, data)
    else:
        return UnknownProject(
            url, name=data["name"], description=data.get("description")
        )


def _create(
    session: Session,
    instance: Instance,
    name: str,
    project_type: str,
    description: Optional[str] = None,
    external_id: Optional[str] = None,
    unified_dataset_name: Optional[str] = None,
) -> Project:
    """Create a project in Tamr.

    Args:
        instance: Tamr instance
        name: Project name
        project_type: Project type
        description: Project description
        external_id: External ID of the project
        unified_dataset_name: Name of the unified dataset

    Returns:
        Project created in Tamr

    Raises:
        project.AlreadyExists: If a project with these specifications already exists.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    if not unified_dataset_name:
        unified_dataset_name = name + "_unified_dataset"
    data = {
        "name": name,
        "type": project_type,
        "unifiedDatasetName": unified_dataset_name,
        "description": description,
        "externalId": external_id,
    }

    project_url = URL(instance=instance, path="projects")
    r = session.post(url=str(project_url), json=data)

    if r.status_code == 409:
        raise AlreadyExists(r.json()["message"])

    data = response.successful(r).json()
    project_path = data["relativeId"]
    project_url = URL(instance=instance, path=str(project_path))

    return _by_url(session=session, url=project_url)


def get_all(
    session: Session,
    instance: Instance,
    *,
    filter: Optional[Union[str, List[str]]] = None,
) -> Tuple[Project, ...]:
    """Get all projects from an instance

    Args:
        instance: Tamr instance from which to get projects
        filter: Filter expression, e.g. "externalId==wobbly"
            Multiple expressions can be passed as a list

    Returns:
        The projects retrieved from the instance

    Raises:
        requests.HTTPError: If an HTTP error is encountered.
    """
    url = URL(instance=instance, path="projects")

    if filter is not None:
        r = session.get(str(url), params={"filter": filter})
    else:
        r = session.get(str(url))

    projects_json = response.successful(r).json()

    projects = []
    for project_json in projects_json:
        project_url = URL(instance=instance, path=project_json["relativeId"])
        project = _from_json(project_url, project_json)
        projects.append(project)
    return tuple(projects)


def attributes(session: Session, project: Project) -> Tuple[Attribute, ...]:
    """Get all attributes from a project

    Args:
        project: Project containing the desired attributes

    Returns:
        The attributes for the specified project

    Raises:
        requests.HTTPError: If an HTTP error is encountered.
    """
    return _get_all_from_parent(session, project)



"""
This module includes:
- SubAttribute
- AttributeType
- Attribute

The definition order is chosen to minimize the number of forward references.
See https://www.python.org/dev/peps/pep-0484/#forward-references

Forward references are necessary because
- `SubAttribute` and `AttributeType` recursively depend on each other
- `Array` and `Map` have `AttributeType` fields but are themselves `AttributeType`s
"""

from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Optional, Tuple, Union

from tamr_client._types.url import URL


# sub attribute
###############


@dataclass(frozen=True)
class SubAttribute:
    """An attribute which is itself a property of another attribute.

    See https://docs.tamr.com/reference#attribute-types

    NOTE:
        `sphinx_autodoc_typehints` cannot handle forward reference to `AttributeType`,
        so reference docs are written manually for this type

    Args:
        name: Name of sub-attribute
        type: See https://docs.tamr.com/reference#attribute-types
        is_nullable: If this sub-attribute can be null
    """

    name: str
    type: "AttributeType"
    is_nullable: bool


# attribute types
#################

# primitive types

PrimitiveType = Enum("PrimitiveType", ["BOOLEAN", "DOUBLE", "INT", "LONG", "STRING"])

# primitive type aliases
DOUBLE = PrimitiveType.DOUBLE
BOOLEAN = PrimitiveType.BOOLEAN
INT = PrimitiveType.INT
LONG = PrimitiveType.LONG
STRING = PrimitiveType.STRING

# complex types


@dataclass(frozen=True)
class Array:
    """See https://docs.tamr.com/reference#attribute-types

    NOTE:
        `sphinx_autodoc_typehints` cannot handle forward reference to `AttributeType`,
        so reference docs are written manually for this type

    Args:
        inner_type
    """

    _tag: ClassVar[str] = "ARRAY"
    inner_type: "AttributeType"


@dataclass(frozen=True)
class Map:
    """See https://docs.tamr.com/reference#attribute-types

    NOTE:
        `sphinx_autodoc_typehints` cannot handle forward reference to `AttributeType`,
        so reference docs are written manually for this type

    Args:
        inner_type
    """

    _tag: ClassVar[str] = "MAP"
    inner_type: "AttributeType"


@dataclass(frozen=True)
class Record:
    """See https://docs.tamr.com/reference#attribute-types

    Args:
        attributes
    """

    _tag: ClassVar[str] = "RECORD"
    attributes: Tuple[SubAttribute, ...]


ComplexType = Union[Array, Map, Record]


AttributeType = Union[PrimitiveType, ComplexType]

# complex type aliases
DEFAULT: AttributeType = Array(STRING)
GEOSPATIAL: AttributeType = Record(
    attributes=(
        SubAttribute(name="point", is_nullable=True, type=Array(DOUBLE)),
        SubAttribute(name="multiPoint", is_nullable=True, type=Array(Array(DOUBLE))),
        SubAttribute(name="lineString", is_nullable=True, type=Array(Array(DOUBLE))),
        SubAttribute(
            name="multiLineString", is_nullable=True, type=Array(Array(Array(DOUBLE)))
        ),
        SubAttribute(
            name="polygon", is_nullable=True, type=Array(Array(Array(DOUBLE)))
        ),
        SubAttribute(
            name="multiPolygon",
            is_nullable=True,
            type=Array(Array(Array(Array(DOUBLE)))),
        ),
    )
)

# attribute
###########


@dataclass(frozen=True)
class Attribute:
    """A Tamr Attribute.

    See https://docs.tamr.com/reference#attribute-types

    Args:
        url
        name
        type
        is_nullable
        description
    """

    url: URL
    name: str
    type: AttributeType
    is_nullable: bool
    description: Optional[str] = None



from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class Instance:
    """Connection parameters for a running Tamr instance

    Args:
        protocol
        host
        port
    """

    protocol: str = "http"
    host: str = "localhost"
    port: Optional[int] = None



from copy import deepcopy

from tamr_unify_client.attribute.type import AttributeType
from tamr_unify_client.base_resource import BaseResource


class Attribute(BaseResource):
    """
    A Tamr Attribute.

    See https://docs.tamr.com/reference#attribute-types
    """

    @classmethod
    def from_json(cls, client, data, api_path):
        return super().from_data(client, data, api_path)

    @property
    def relative_id(self):
        """:type: str"""
        # api_path is alias when it exists, and relative_id when it does not.
        # this distinction is useful for things like refreshing a unified dataset,
        # where using the relative_id would hit
        #   /datasets/{id}:refresh
        # rather than
        #   /projects/{id}/unifiedDataset:refresh.
        # Since attributes don't currently have that kind of aliasing,
        # using api_path is always correct.
        # If attributes ever get aliased, this will need to be updated.
        # This is confusing; there's an RFC for suggestions to improve this
        # #64 https://github.com/Datatamer/unify-client-python/issues/64
        # "Conflation between 'api_path', 'relative_id' / 'relativeId', and
        # BaseResource ctor 'alias'"
        return self.api_path

    @property
    def name(self):
        """:type: str"""
        return self._data.get("name")

    @property
    def description(self):
        """:type: str"""
        return self._data.get("description")

    @property
    def type(self):
        """:type: :class:`~tamr_unify_client.attribute.type.AttributeType`"""
        type_json = self._data.get("type")
        return AttributeType(type_json)

    @property
    def is_nullable(self):
        """:type: bool"""
        return self._data.get("isNullable")

    def spec(self):
        """Returns a spec representation of this attribute.

        :return: The attribute spec.
        :rtype: :class:`~tamr_unify_client.attribute.resource.AttributeSpec`
        """
        return AttributeSpec.of(self)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"relative_id={self.relative_id!r}, "
            f"name={self.name!r})"
        )


class AttributeSpec:
    """A representation of the server view of an attribute"""

    def __init__(self, client, data, api_path):
        self._data = data
        self.client = client
        self.api_path = api_path

    @staticmethod
    def of(resource):
        """Creates an attribute spec from an attribute.

        :param resource: The existing attribute.
        :type resource: :class:`~tamr_unify_client.attribute.resource.Attribute`
        :return: The corresponding attribute spec.
        :rtype: :class:`~tamr_unify_client.attribute.resource.AttributeSpec`
        """
        return AttributeSpec(
            resource.client, deepcopy(resource._data), resource.api_path
        )

    @staticmethod
    def new():
        """Creates a blank spec that could be used to construct a new attribute.

        :return: The empty spec.
        :rtype: :class:`~tamr_unify_client.attribute.resource.AttributeSpec`
        """
        return AttributeSpec(None, {}, None)

    def from_data(self, data):
        """Creates a spec with the same client and API path as this one, but new data.

        :param data: The data for the new spec.
        :type data: dict
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.attribute.resource.AttributeSpec`
        """
        return AttributeSpec(self.client, data, self.api_path)

    def to_dict(self):
        """Returns a version of this spec that conforms to the API representation.

        :returns: The spec's dict.
        :rtype: dict
        """
        return deepcopy(self._data)

    def with_name(self, new_name):
        """Creates a new spec with the same properties, updating name.

        :param new_name: The new name.
        :type new_name: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.attribute.resource.AttributeSpec`
        """
        return self.from_data({**self._data, "name": new_name})

    def with_description(self, new_description):
        """Creates a new spec with the same properties, updating description.

        :param new_description: The new description.
        :type new_description: str
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.attribute.resource.AttributeSpec`
        """
        return self.from_data({**self._data, "description": new_description})

    def with_type(self, new_type):
        """Creates a new spec with the same properties, updating type.

        :param new_type: The spec of the new type.
        :type new_type: :class:`~tamr_unify_client.attribute.type.AttributeTypeSpec`
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.attribute.resource.AttributeSpec`
        """
        type_spec = new_type.to_dict()
        return self.from_data({**self._data, "type": type_spec})

    def with_is_nullable(self, new_is_nullable):
        """Creates a new spec with the same properties, updating is nullable.

        :param new_is_nullable: The new is nullable.
        :type new_is_nullable: bool
        :return: The new spec.
        :rtype: :class:`~tamr_unify_client.attribute.resource.AttributeSpec`
        """
        return self.from_data({**self._data, "isNullable": new_is_nullable})

    def put(self):
        """Commits the changes and updates the attribute in Tamr.

        :return: The updated attribute.
        :rtype: :class:`~tamr_unify_client.attribute.resource.Attribute`
        """
        updated_attribute = (
            self.client.put(self.api_path, json=self._data).successful().json()
        )
        return Attribute.from_json(self.client, updated_attribute, self.api_path)

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"dict={self._data!r})"
        )



"""
Tamr - Schema Mapping
See https://docs.tamr.com/new/docs/overall-workflow-schema
"""
from tamr_client.schema_mapping import attribute_mapping, project
from tamr_client.schema_mapping._schema_mapping import update_unified_dataset



import warnings

from requests.exceptions import HTTPError

from tamr_unify_client.base_collection import BaseCollection
from tamr_unify_client.dataset.resource import Dataset


class DatasetCollection(BaseCollection):
    """Collection of :class:`~tamr_unify_client.dataset.resource.Dataset` s.

    :param client: Client for API call delegation.
    :type client: :class:`~tamr_unify_client.Client`
    :param api_path: API path used to access this collection.
        E.g. ``"projects/1/inputDatasets"``.
        Default: ``"datasets"``.
    :type api_path: str
    """

    def __init__(self, client, api_path="datasets"):
        super().__init__(client, api_path)

    def by_resource_id(self, resource_id):
        """Retrieve a dataset by resource ID.

        :param resource_id: The resource ID. E.g. ``"1"``
        :type resource_id: str
        :returns: The specified dataset.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """
        return super().by_resource_id("datasets", resource_id)

    def by_relative_id(self, relative_id):
        """Retrieve a dataset by relative ID.

        :param relative_id: The resource ID. E.g. ``"datasets/1"``
        :type relative_id: str
        :returns: The specified dataset.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """
        return super().by_relative_id(Dataset, relative_id)

    def by_external_id(self, external_id):
        """Retrieve a dataset by external ID.

        :param external_id: The external ID.
        :type external_id: str
        :returns: The specified dataset, if found.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        :raises KeyError: If no dataset with the specified external_id is found
        :raises LookupError: If multiple datasets with the specified external_id are found
        """
        return super().by_external_id(Dataset, external_id)

    def stream(self):
        """Stream datasets in this collection. Implicitly called when iterating
        over this collection.

        :returns: Stream of datasets.
        :rtype: Python generator yielding :class:`~tamr_unify_client.dataset.resource.Dataset`

        Usage:
            >>> for dataset in collection.stream(): # explicit
            >>>     do_stuff(dataset)
            >>> for dataset in collection: # implicit
            >>>     do_stuff(dataset)
        """
        return super().stream(Dataset)

    def by_name(self, dataset_name):
        """Lookup a specific dataset in this collection by exact-match on name.

        :param dataset_name: Name of the desired dataset.
        :type dataset_name: str
        :return: Dataset with matching name in this collection.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        :raises KeyError: If no dataset with specified name was found.
        """
        for dataset in self:
            if dataset.name == dataset_name:
                return dataset
        raise KeyError(f"No dataset found with name: {dataset_name}")

    def delete_by_resource_id(self, resource_id, cascade=False):
        """Deletes a dataset from this collection by resource_id. Optionally deletes all derived datasets as well.

        :param resource_id: The resource id of the dataset in this collection to delete.
        :type resource_id: str
        :param cascade: Whether to delete all datasets derived from the deleted one. Optional, default is `False`.
            Do not use this option unless you are certain you need it as it can have unindended consequences.
        :type cascade: bool
        :return: HTTP response from the server.
        :rtype: :class:`requests.Response`
        """
        params = {"cascade": cascade}
        path = f"{self.api_path}/{resource_id}"
        response = self.client.delete(path, params=params).successful()
        return response

    def create(self, creation_spec):
        """
        Create a Dataset in Tamr

        :param creation_spec: Dataset creation specification should be formatted as specified in the `Public Docs for Creating a Dataset <https://docs.tamr.com/reference#create-a-dataset>`_.
        :type creation_spec: dict[str, str]
        :returns: The created Dataset
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        """
        data = self.client.post(self.api_path, json=creation_spec).successful().json()
        return Dataset.from_json(self.client, data)

    def create_from_dataframe(
        self, df, primary_key_name, dataset_name, ignore_nan=None
    ):
        """Creates a dataset in this collection with the given name, creates an attribute for each column in the `df`
        (with `primary_key_name` as the key attribute), and upserts a record for each row of `df`.

        Each attribute has the default type `ARRAY[STRING]`, besides the key attribute, which will have type `STRING`.

        This function attempts to ensure atomicity, but it is not guaranteed. If an error occurs while creating
        attributes or records, an attempt will be made to delete the dataset that was created. However, if this
        request errors, it will not try again.

        :param df: The data to create the dataset with.
        :type df: :class:`pandas.DataFrame`
        :param primary_key_name: The name of the primary key of the dataset. Must be a column of `df`.
        :type primary_key_name: str
        :param dataset_name: What to name the dataset in Tamr. There cannot already be a dataset with this name.
        :type dataset_name: str
        :param ignore_nan: Legacy parameter that does nothing
        :type ignore_nan: bool
        :returns: The newly created dataset.
        :rtype: :class:`~tamr_unify_client.dataset.resource.Dataset`
        :raises KeyError: If `primary_key_name` is not a column in `df`.
        :raises CreationError: If a step in creating the dataset fails.
        """
        if ignore_nan is not None:
            warnings.warn(
                "'ignore_nan' is deprecated. DataFrame `NaN`s are always ignored in upsert",
                DeprecationWarning,
            )
        if primary_key_name not in df.columns:
            raise KeyError(f"{primary_key_name} is not an attribute of the data")

        creation_spec = {"name": dataset_name, "keyAttributeNames": [primary_key_name]}
        try:
            dataset = self.create(creation_spec)
        except HTTPError:
            raise CreationError("Dataset was not created")
        # after this point, if a request fails, try to undo the change by deleting this dataset

        attributes = dataset.attributes
        for col in df.columns:
            if col == primary_key_name:
                # this attribute already exists, so don't create it again
                continue

            attr_spec = {
                "name": col,
                "type": {"baseType": "ARRAY", "innerType": {"baseType": "STRING"}},
            }
            try:
                attributes.create(attr_spec)
            except HTTPError:
                self._handle_creation_failure(dataset, "An attribute was not created")

        try:
            response = dataset.upsert_from_dataframe(
                df, primary_key_name=primary_key_name
            )
        except HTTPError:
            self._handle_creation_failure(dataset, "Records could not be created")

        if not response["allCommandsSucceeded"]:
            self._handle_creation_failure(dataset, "Some records had validation errors")

        return dataset

    def _handle_creation_failure(self, dataset, error):
        """Attempts to make create_from_dataframe atomic by deleting the created dataset in the event of later failure.
        However, this does not guarantee atomicity: if the request to delete the dataset fails, it will not retry.

        :param dataset: The created dataset to delete.
        :type dataset: :class:`~tamr_unify_client.dataset.resource.Dataset`
        :param error: The error that caused the function to fail.
        :type error: str
        """
        try:
            dataset.delete()
        except HTTPError:
            raise CreationError("Created dataset didn't delete after an earlier error")
        raise CreationError(error)

    # super.__repr__ is sufficient


class CreationError(Exception):
    """An error from :func:`~tamr_unify_client.dataset.collection.DatasetCollection.create_from_dataframe`"""

    def __init__(self, error_message):
        super().__init__(error_message)



"""
See https://docs.tamr.com/new/reference/the-operation-object
"""
from copy import deepcopy
from time import sleep, time as now
from typing import Optional

import requests

from tamr_client import response
from tamr_client._types import Instance, JsonDict, Operation, Session, URL
from tamr_client.exception import TamrClientException


class NotFound(TamrClientException):
    """Raised when referencing an operation that does not exist on the server.
    """

    pass


class Failed(TamrClientException):
    """Raised when checking a failed operation.
    """

    pass


def check(session: Session, operation: Operation):
    """Waits for the operation to finish and raises an exception if the operation was not successful.

    Args:
        operation: Operation to be checked.

    Raises:
        Failed: If the operation failed.
    """
    op = wait(session, operation)
    if not succeeded(op):
        raise Failed(
            f"Checked operation '{str(op.url)}', but it failed with status: {op.status}"
        )


def poll(session: Session, operation: Operation) -> Operation:
    """Poll this operation for server-side updates.

    Does not update the :class:`~tamr_client.operation.Operation` object.
    Instead, returns a new :class:`~tamr_client.operation.Operation`.

    Args:
        operation: Operation to be polled.
    """
    return _by_url(session, operation.url)


def wait(
    session: Session,
    operation: Operation,
    *,
    poll_interval_seconds: int = 3,
    timeout_seconds: Optional[int] = None,
) -> Operation:
    """Continuously polls for this operation's server-side state.

    Args:
        operation: Operation to be polled.
        poll_interval_seconds: Time interval (in seconds) between subsequent polls.
        timeout_seconds: Time (in seconds) to wait for operation to resolve.

    Raises:
        TimeoutError: If operation takes longer than `timeout_seconds` to resolve.
    """
    started = now()
    while timeout_seconds is None or now() - started < timeout_seconds:
        if operation.status is None:
            return operation
        elif operation.status["state"] in ["PENDING", "RUNNING"]:
            sleep(poll_interval_seconds)
        elif operation.status["state"] in ["CANCELED", "SUCCEEDED", "FAILED"]:
            return operation
        operation = poll(session, operation)
    raise TimeoutError(
        f"Waiting for operation took longer than {timeout_seconds} seconds."
    )


def succeeded(operation: Operation) -> bool:
    """Convenience method for checking if operation was successful.
    """
    return operation.status is not None and operation.status["state"] == "SUCCEEDED"


def by_resource_id(session: Session, instance: Instance, resource_id: str) -> Operation:
    """Get operation by ID

    Args:
        resource_id: The ID of the operation
    """
    url = URL(instance=instance, path=f"operations/{resource_id}")
    r = session.get(str(url))
    return _from_response(instance, r)


def _from_response(instance: Instance, response: requests.Response) -> Operation:
    """
    Handle idiosyncrasies in constructing Operations from Tamr responses.
    When a Tamr API call would start an operation, but all results that would be
    produced by that operation are already up-to-date, Tamr returns `HTTP 204 No Content`

    To make it easy for client code to handle these API responses without checking
    the response code, this method will either construct an Operation, or a
    dummy `NoOp` operation representing the 204 Success response.

    Args:
        response: HTTP Response from the request that started the operation.
    """
    if response.status_code == 204:
        # Operation was successful, but the response contains no content.
        # Create a dummy operation to represent this.
        _never = "0000-00-00T00:00:00.000Z"
        _description = """Tamr returned HTTP 204 for this operation, indicating that all
            results that would be produced by the operation are already up-to-date."""
        resource_json = {
            "id": "-1",
            "type": "NOOP",
            "description": _description,
            "status": {
                "state": "SUCCEEDED",
                "startTime": _never,
                "endTime": _never,
                "message": "",
            },
            "created": {"username": "", "time": _never, "version": "-1"},
            "lastModified": {"username": "", "time": _never, "version": "-1"},
            "relativeId": "operations/-1",
        }
    else:
        resource_json = response.json()
    _id = resource_json["id"]
    _url = URL(instance=instance, path=f"operations/{_id}")
    return _from_json(_url, resource_json)


def _by_url(session: Session, url: URL) -> Operation:
    """Get operation by URL

    Fetches operation from Tamr server

    Args:
        url: Operation URL

    Raises:
        operation.NotFound: If no operation could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    r = session.get(str(url))
    if r.status_code == 404:
        raise NotFound(str(url))
    data = response.successful(r).json()
    return _from_json(url, data)


def _from_json(url: URL, data: JsonDict) -> Operation:
    """Make operation from JSON data (deserialize)

    Args:
        url: Operation URL
        data: Operation JSON data from Tamr server
    """
    cp = deepcopy(data)
    return Operation(
        url, type=cp["type"], status=cp.get("status"), description=cp.get("description")
    )



from tamr_unify_client.base_resource import BaseResource
from tamr_unify_client.operation import Operation


class EstimatedPairCounts(BaseResource):
    """Estimated Pair Counts info for Mastering Project"""

    @classmethod
    def from_json(cls, client, resource_json, api_path=None) -> "EstimatedPairCounts":
        return super().from_data(client, resource_json, api_path)

    @property
    def is_up_to_date(self) -> bool:
        """Whether an estimate pairs job has been run since the last edit to the binning model.

        :rtype: bool
        """
        return self._data.get("isUpToDate")

    @property
    def total_estimate(self) -> dict:
        """The total number of estimated candidate pairs and generated pairs for the model across all clauses.

        :return: A dictionary containing candidate pairs and estimated pairs mapped to their corresponding estimated counts.
            For example:\n
            {\n
                "candidatePairCount": "54321",\n
                "generatedPairCount": "12345"\n
            }
        :rtype: dict[str, str]
        """
        return self._data.get("totalEstimate")

    @property
    def clause_estimates(self) -> dict:
        """The estimated candidate pair count and generated pair count for each clause in the model.

        :return: A dictionary containing each clause name mapped to a dictionary containing the corresponding estimated candidate and generated pair counts.
            For example:\n
            {\n
                "Clause1": {\n
                    "candidatePairCount": "321",\n
                    "generatedPairCount": "123"\n
                },\n
                "Clause2": {\n
                    "candidatePairCount": "654",\n
                    "generatedPairCount": "456"\n
                }\n
            }
        :rtype: dict[str, dict[str, str]]
        """
        return self._data.get("clauseEstimates")

    def refresh(self, **options):
        """Updates the estimated pair counts if needed.

        The pair count estimates are updated on the server; you will need to call
        :func:`~tamr_unify_client.mastering.project.MasteringProject.estimate_pairs`
        to retrieve the updated estimate.

        :param ``**options``: Options passed to underlying :class:`~tamr_unify_client.operation.Operation` .
            See :func:`~tamr_unify_client.operation.Operation.apply_options` .
        :returns: The refresh operation.
        :rtype: :class:`~tamr_unify_client.operation.Operation`
        """
        response = self.client.post(self.api_path + ":refresh").successful()
        op = Operation.from_response(self.client, response)
        return op.apply_options(**options)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"up_to_date={self.is_up_to_date!r}, "
            f"total_estimate={self.total_estimate!r}, "
            f"clause_estimates={self.clause_estimates!r})"
        )



import os
import sys


def check():
    env_var = "TAMR_CLIENT_BETA"
    is_beta_enabled = os.environ.get(env_var) == "1"

    if not is_beta_enabled:
        msg = (
            f"ERROR: 'tamr_client' package is in BETA, but you do not have the '{env_var}' environment variable set to '1'."
            "\n\nHINT: For non-BETA features, use only the 'tamr_unify_client' package."
            f"\nHINT: To opt-in to BETA features, set environment variable: '{env_var}=1'."
            "\n\nWARNING: Do not rely on BETA features in production workflows."
            " Support from Tamr may be limited."
        )
        print(msg)
        sys.exit(1)



import logging

import requests

logger = logging.getLogger(__name__)


def successful(response: requests.Response) -> requests.Response:
    """Ensure response does not contain an HTTP error.

    Delegates to :func:`requests.Response.raise_for_status`

    Returns:
        The response being checked.

    Raises:
        requests.exceptions.HTTPError: If an HTTP error is encountered.
    """
    try:
        response.raise_for_status()
    except requests.HTTPError as e:
        r = e.response
        logger.error(
            f"Encountered HTTP error code {r.status_code}. Response body: {r.text}"
        )
        raise e
    return response


def _monkey_patch():
    requests.Response.successful = successful



class ProjectStep:
    """A step of a Tamr project. This is not a `BaseResource` because it has no API path
    and cannot be directly retrieved or modified.

    See https://docs.tamr.com/reference#retrieve-downstream-dataset-usage

    :param client: Delegate underlying API calls to this client.
    :type client: :class:`~tamr_unify_client.Client`
    :param data: The JSON body containing project step information.
    :type data: :py:class:`dict`
    """

    def __init__(self, client, data):
        self.client = client
        self._data = data

    @property
    def project_step_id(self):
        """:type: str"""
        return self._data.get("projectStepId")

    @property
    def project_step_name(self):
        """:type: str"""
        return self._data.get("projectStepName")

    @property
    def project_name(self):
        """:type: str"""
        return self._data.get("projectName")

    @property
    def type(self):
        """A Tamr project type, listed in https://docs.tamr.com/reference#create-a-project.

        :type: str"""
        return self._data.get("type")

    def project(self):
        """Retrieves the :class:`~tamr_unify_client.project.resource.Project` this step is associated with.

        :returns: This step's project.
        :rtype: :class:`~tamr_unify_client.project.resource.Project`
        :raises KeyError: If no project with the specified name is found.
        :raises LookupError: If multiple projects with the specified name are found.
        """
        name = self.project_name
        projects = [p for p in self.client.projects if p.name == name]

        if len(projects) == 0:
            raise KeyError(f'No project found with name "{name}"')
        elif len(projects) > 1:
            raise LookupError(f'Multiple projects found with name "{name}"')

        return projects[0]

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"project_step_id={self.project_step_id!r}, "
            f"project_step_name={self.project_step_name!r}, "
            f"project_name={self.project_name!r}, "
            f"type={self.type!r})"
        )



from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from tamr_unify_client.attribute.type import AttributeType

SubAttributeJson = Dict[str, Any]


@dataclass(frozen=True)
class SubAttribute:
    """An attribute which is itself a property of another attribute.

    See https://docs.tamr.com/reference#attribute-types

    Args:
        name: Name of sub-attribute
        description: Description of sub-attribute
        type: See https://docs.tamr.com/reference#attribute-types
        is_nullable: If this sub-attribute can be null
    """

    name: str
    type: AttributeType
    is_nullable: bool
    _json: SubAttributeJson = field(repr=False)
    description: Optional[str] = None

    @staticmethod
    def from_json(data: SubAttributeJson) -> "SubAttribute":
        """Create a SubAttribute from JSON data.

        Args:
            data: JSON data received from Tamr server.
        """
        _json = deepcopy(data)

        dc = deepcopy(data)
        dc["is_nullable"] = dc.pop("isNullable")

        type_json = dc.pop("type")
        # TODO implement AttributeType.from_json and use that instead
        type = AttributeType(type_json)

        return SubAttribute(**dc, type=type, _json=_json)



"""
See https://docs.tamr.com/new/reference/retrieve-projects-mappings
"""
from copy import deepcopy
from typing import Dict, List, Optional, Tuple

from tamr_client import attribute, dataset, response
from tamr_client import project as tc_project
from tamr_client._types import (
    Attribute,
    AttributeMapping,
    Instance,
    JsonDict,
    Project,
    Session,
    URL,
)
from tamr_client.exception import TamrClientException


class NotFound(TamrClientException):
    """Raised when referencing an attribute mapping that does not exist on the server."""

    pass


class AlreadyExists(TamrClientException):
    """Raised when an attribute mapping with these specifications already exists."""

    pass


class Ambiguous(TamrClientException):
    """Raised when an attribute mapping specification is incomplete, ambiguous, or contradictory."""

    pass


def create(
    session: Session,
    project: Project,
    input_attribute: Attribute,
    unified_attribute: Attribute,
) -> AttributeMapping:
    """Create a mapping in Tamr between input attributes and unified attributes of the given project.

    Args:
        project: Tamr project
        input_attribute: The attribute of a source dataset to map
        unified_attribute: The attribute of the unified dataset to map onto

    Returns:
        Attribute mapping created in Tamr

    Raises:
        attribute_mapping.AlreadyExists: If an attribute mapping with these specifications already exists.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    data = {
        "relativeInputAttributeId": input_attribute.url.path,
        "relativeUnifiedAttributeId": unified_attribute.url.path,
    }
    r = session.post(url=str(project.url) + "/attributeMappings", json=data)

    if r.status_code == 204:
        raise AlreadyExists(
            "An attribute mapping with these specifications already exists."
        )
    if r.status_code == 400:
        raise Ambiguous(r.json()["message"])
    if r.status_code == 404:
        if "dataset" in r.json()["message"].lower():
            raise dataset.NotFound(r.json()["message"])
        elif "project" in r.json()["message"].lower():
            raise tc_project.NotFound(r.json()["message"])

    data = response.successful(r).json()
    instance = project.url.instance
    url = URL(instance=instance, path=data["relativeId"])
    return AttributeMapping(
        url,
        input_attribute=attribute._attribute._by_url(
            session, URL(instance=instance, path=data["relativeInputAttributeId"]),
        ),
        unified_attribute=attribute._attribute._by_url(
            session, URL(instance=instance, path=data["relativeUnifiedAttributeId"]),
        ),
    )


def get_all(session: Session, tamr_project: Project) -> List[AttributeMapping]:
    """Get all attribute mappings of a Tamr project

    Args:
        tamr_project: Tamr project

    Returns:
        The attribute mappings of the project

    Raises:
        requests.HTTPError: If an HTTP error is encountered.
    """
    url = str(tamr_project.url) + "/attributeMappings"
    r = session.get(url=url)

    data = response.successful(r).json()
    mapping_list = []
    attribute_memo: Dict[URL, Attribute] = {}
    for mapping_data in data:
        mapping, attribute_memo = _get(
            session,
            tamr_project.url.instance,
            mapping_data,
            attribute_memo=attribute_memo,
        )
        mapping_list.append(mapping)
    return mapping_list


def _get(
    session: Session,
    instance: Instance,
    data: JsonDict,
    *,
    attribute_memo: Optional[Dict[URL, Attribute]] = None
) -> Tuple[AttributeMapping, Dict[URL, Attribute]]:
    """Construct an attribute mapping from server response

    This is a utility function used by `attribute_mapping.get_all`. When an attribute is
    fetched, all other attributes in its dataset are also retrieved and stored in a
    dictionary to limit the number of requests made to the server

    Args:
        instance: A Tamr instance
        data: The JSON body of the server response retrieve a project's attribute mappings
        attribute_memo: Dictionary of fetched attributes used to avoid redundant requests

    Returns:
        The attribute mapping and updated memo
    """
    if attribute_memo is None:
        attribute_memo = {}
    cp = deepcopy(data)
    input_attribute_url = URL(instance=instance, path=cp["relativeInputAttributeId"])
    unified_attribute_url = URL(
        instance=instance, path=cp["relativeUnifiedAttributeId"]
    )

    if input_attribute_url not in attribute_memo:
        attr_dataset = dataset.by_resource_id(
            session, instance, cp["relativeInputAttributeId"].split("/")[1]
        )
        attribute_memo = {
            **attribute_memo,
            **{attr.url: attr for attr in dataset.attributes(session, attr_dataset)},
        }
    if unified_attribute_url not in attribute_memo:
        attr_dataset = dataset.by_resource_id(
            session, instance, cp["relativeUnifiedAttributeId"].split("/")[1]
        )
        attribute_memo = {
            **attribute_memo,
            **{attr.url: attr for attr in dataset.attributes(session, attr_dataset)},
        }

    return (
        AttributeMapping(
            url=URL(instance=instance, path=cp["relativeId"]),
            input_attribute=attribute_memo[input_attribute_url],
            unified_attribute=attribute_memo[unified_attribute_url],
        ),
        attribute_memo,
    )


def delete(session: Session, attribute_mapping: AttributeMapping):
    """Delete an existing attribute mapping

    Args:
        attribute_mapping: Existing attribute mapping to delete

    Raises:
        attribute_mapping.NotFound: If no attribute mapping could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    r = session.delete(str(attribute_mapping.url))
    if r.status_code == 404:
        raise NotFound(str(attribute_mapping.url))
    response.successful(r)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tamr_client',
 'tamr_client._types',
 'tamr_client.attribute',
 'tamr_client.categorization',
 'tamr_client.dataset',
 'tamr_client.golden_records',
 'tamr_client.mastering',
 'tamr_client.schema_mapping',
 'tamr_unify_client',
 'tamr_unify_client.attribute',
 'tamr_unify_client.auth',
 'tamr_unify_client.categorization',
 'tamr_unify_client.categorization.category',
 'tamr_unify_client.dataset',
 'tamr_unify_client.mastering',
 'tamr_unify_client.mastering.published_cluster',
 'tamr_unify_client.project',
 'tamr_unify_client.project.attribute_configuration',
 'tamr_unify_client.project.attribute_mapping']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.22,<3.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses>=0.6.0,<0.7.0']}

setup_kwargs = {
    'name': 'tamr-unify-client',
    'version': '1.5.0',
    'description': 'Python Client for the Tamr API',
    'long_description': "# Python Client\nProgrammatically 💻 interact with Tamr using Python 🐍\n\n[![Version](https://img.shields.io/pypi/v/tamr-unify-client.svg?style=flat-square)](https://pypi.org/project/tamr-unify-client/)\n[![Documentation Status](https://readthedocs.org/projects/tamr-client/badge/?version=stable&style=flat-square)](https://tamr-client.readthedocs.io/en/stable/?badge=stable)\n[![Build Status](https://img.shields.io/github/workflow/status/Datatamer/tamr-client/CI?&style=flat-square)](https://github.com/Datatamer/tamr-client/actions?query=workflow%3ACI)\n![Supported Python Versions](https://img.shields.io/pypi/pyversions/tamr-unify-client.svg?style=flat-square)\n[![License](https://img.shields.io/pypi/l/tamr-unify-client.svg?style=flat-square)](LICENSE)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square)](https://github.com/ambv/black)\n\n---\n\n*Quick links:*\n**[Docs](https://tamr-client.readthedocs.io/en/stable/)** |\n**[Contributing](https://tamr-client.readthedocs.io/en/stable/contributor-guide.html)** |\n**[License](https://github.com/Datatamer/tamr-client/blob/main/LICENSE)**\n\n---\n\n## Install\n\n```sh\npip install tamr-unify-client\n```\n\n## Features\n- 🐍 Python objects/methods/functions instead of raw HTTP requests\n- 🤖 Automate operational workflows\n  - Continuous Mastering\n  - Continuous Categorization\n- 🚀 Kick-off synchronous/asynchronous operations\n  - Refresh datasets in your pipeline\n  - Train Tamr's machine learning models\n  - Generate predictions from trained models\n- 🔒 Authenticate with Tamr\n\nFor more see the [official docs](https://tamr-client.readthedocs.io/en/stable/).\n\n## Maintainers\n\n- [Pedro Cattori](https://github.com/pcattori)\n- [Samuel Kalish](https://github.com/skalish)\n",
    'author': 'Pedro Cattori',
    'author_email': 'pedro.cattori@tamr.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://tamr-client.readthedocs.io/en/stable/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



from typing import Optional

from tamr_client import project
from tamr_client._types import (
    CategorizationProject,
    Instance,
    JsonDict,
    Project,
    Session,
    URL,
)


def _from_json(url: URL, data: JsonDict) -> CategorizationProject:
    """Make Categorization project from JSON data (deserialize)

    Args:
        url: Project URL
        data: Project JSON data from Tamr server
    """
    return CategorizationProject(
        url, name=data["name"], description=data.get("description")
    )


def create(
    session: Session,
    instance: Instance,
    name: str,
    description: Optional[str] = None,
    external_id: Optional[str] = None,
    unified_dataset_name: Optional[str] = None,
) -> Project:
    """Create a Categorization project in Tamr.

    Args:
        instance: Tamr instance
        name: Project name
        description: Project description
        external_id: External ID of the project
        unified_dataset_name: Unified dataset name. If None, will be set to project name + _'unified_dataset'

    Returns:
        Project created in Tamr

    Raises:
        project.AlreadyExists: If a project with these specifications already exists
        requests.HTTPError: If any other HTTP error is encountered
    """
    return project._create(
        session=session,
        instance=instance,
        name=name,
        project_type="CATEGORIZATION",
        description=description,
        external_id=external_id,
        unified_dataset_name=unified_dataset_name,
    )



"""
See https://docs.tamr.com/reference/attribute-types
"""
from copy import deepcopy
from dataclasses import replace
from typing import Optional, Tuple, Union

from tamr_client import response
from tamr_client._types import (
    Attribute,
    AttributeType,
    Dataset,
    JsonDict,
    Project,
    Session,
    UnifiedDataset,
    URL,
)
from tamr_client.attribute import type as attribute_type
from tamr_client.exception import TamrClientException


_RESERVED_NAMES = frozenset(
    [
        # See javasrc/procurify/ui/app/scripts/constants/ElasticConstants.js
        "origin_source_name",
        "tamr_id",
        "origin_entity_id",
        # See javasrc/procurify/ui/app/scripts/constants/PipelineConstants.js
        "clusterId",
        "originSourceId",
        "originEntityId",
        "sourceId",
        "entityId",
        "suggestedClusterId",
        "verificationType",
        "verifiedClusterId",
    ]
)

Parent = Union[Dataset, Project]


class AlreadyExists(TamrClientException):
    """Raised when trying to create an attribute that already exists on the server"""

    pass


class NotFound(TamrClientException):
    """Raised when referencing (e.g. updating or deleting) an attribute
    that does not exist on the server.
    """

    pass


class ReservedName(TamrClientException):
    """Raised when attempting to create an attribute with a reserved name"""

    pass


class CannotCreateAttributesOnUnifiedDataset(TamrClientException):
    """Raised when attempting to create an attribute on a unified dataset"""

    pass


def by_resource_id(
    session: Session, parent: Union[Dataset, Project], id: str
) -> Attribute:
    """Get attribute by resource ID

    Fetches attribute from Tamr server

    Args:
        parent: Dataset or project containing this attribute
        id: Attribute ID

    Raises:
        attribute.NotFound: If no attribute could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    url = replace(parent.url, path=parent.url.path + f"/attributes/{id}")
    return _by_url(session, url)


def _by_url(session: Session, url: URL) -> Attribute:
    """Get attribute by URL

    Fetches attribute from Tamr server

    Args:
        url: Attribute URL

    Raises:
        attribute.NotFound: If no attribute could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    r = session.get(str(url))
    if r.status_code == 404:
        raise NotFound(str(url))
    data = response.successful(r).json()
    return _from_json(url, data)


def _from_json(url: URL, data: JsonDict) -> Attribute:
    """Make attribute from JSON data (deserialize)

    Args:
        url: Attribute URL
        data: Attribute JSON data from Tamr server
    """
    cp = deepcopy(data)
    return Attribute(
        url,
        name=cp["name"],
        description=cp.get("description"),
        is_nullable=cp["isNullable"],
        type=attribute_type.from_json(cp["type"]),
    )


def to_json(attr: Attribute) -> JsonDict:
    """Serialize attribute into JSON

    Args:
        attr: Attribute to serialize

    Returns:
        JSON data representing the attribute
    """
    d = {
        "name": attr.name,
        "type": attribute_type.to_json(attr.type),
        "isNullable": attr.is_nullable,
    }
    if attr.description is not None:
        d["description"] = attr.description
    return d


def create(
    session: Session,
    parent: Parent,
    *,
    name: str,
    is_nullable: bool,
    type: AttributeType = attribute_type.DEFAULT,
    description: Optional[str] = None,
) -> Attribute:
    """Create an attribute

    Posts a creation request to the Tamr server

    Args:
        parent: Dataset or project that should contain the new attribute
        name: Name for the new attribute
        type: Attribute type for the new attribute
        is_nullable: Determines if the new attribute can contain NULL values
        description: Description of the new attribute
        force: If `True`, skips reserved attribute name check

    Returns:
        The newly created attribute

    Raises:
        attribute.ReservedName: If attribute name is reserved.
        attribute.AlreadyExists: If an attribute already exists at the specified URL.
            Corresponds to a 409 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    if name in _RESERVED_NAMES:
        raise ReservedName(name)

    if isinstance(parent, UnifiedDataset):
        raise CannotCreateAttributesOnUnifiedDataset(
            "Attributes for unified datasets must be created as attributes of the "
            "containing project"
        )

    return _create(
        session,
        parent,
        name=name,
        is_nullable=is_nullable,
        type=type,
        description=description,
    )


def _create(
    session: Session,
    parent: Parent,
    *,
    name: str,
    is_nullable: bool,
    type: AttributeType = attribute_type.DEFAULT,
    description: Optional[str] = None,
) -> Attribute:
    """Same as `tc.attribute.create`, but does not check for reserved attribute
    names.
    """
    attrs_url = replace(parent.url, path=parent.url.path + "/attributes")
    url = replace(attrs_url, path=attrs_url.path + f"/{name}")

    body = {
        "name": name,
        "type": attribute_type.to_json(type),
        "isNullable": is_nullable,
    }
    if description is not None:
        body["description"] = description

    r = session.post(str(attrs_url), json=body)
    if r.status_code == 409:
        raise AlreadyExists(str(url))
    data = response.successful(r).json()

    return _from_json(url, data)


def update(
    session: Session, attribute: Attribute, *, description: Optional[str] = None
) -> Attribute:
    """Update an existing attribute

    PUTS an update request to the Tamr server

    Args:
        attribute: Existing attribute to update
        description: Updated description for the existing attribute

    Returns:
        The newly updated attribute

    Raises:
        attribute.NotFound: If no attribute could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    updates = {"description": description}
    r = session.put(str(attribute.url), json=updates)
    if r.status_code == 404:
        raise NotFound(str(attribute.url))
    data = response.successful(r).json()
    return _from_json(attribute.url, data)


def delete(session: Session, attribute: Attribute):
    """Deletes an existing attribute

    Sends a deletion request to the Tamr server

    Args:
        attribute: Existing attribute to delete

    Raises:
        attribute.NotFound: If no attribute could be found at the specified URL.
            Corresponds to a 404 HTTP error.
        requests.HTTPError: If any other HTTP error is encountered.
    """
    r = session.delete(str(attribute.url))
    if r.status_code == 404:
        raise NotFound(str(attribute.url))
    response.successful(r)


def _get_all_from_parent(session: Session, parent: Parent) -> Tuple[Attribute, ...]:
    """Get all attributes belonging to a parent entity

    Args:
        parent: Entity to fetch attributes from

    Returns:
        The attributes for the specified parent

    Raises:
        requests.HTTPError: If an HTTP error is encountered.
    """
    attrs_url = replace(parent.url, path=parent.url.path + "/attributes")
    r = session.get(str(attrs_url))
    attrs_json = response.successful(r).json()

    def attr_from_json(attr_json: JsonDict) -> Attribute:
        attr_url = replace(attrs_url, path=attrs_url.path + f'/{attr_json["name"]}')
        return _from_json(attr_url, attr_json)

    return tuple(attr_from_json(attr_json) for attr_json in attrs_json)



from tamr_unify_client.base_collection import BaseCollection
from tamr_unify_client.project.attribute_configuration.resource import (
    AttributeConfiguration,
)


class AttributeConfigurationCollection(BaseCollection):
    """Collection of :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfiguration`

    :param client: Client for API call delegation.
    :type client: :class:`~tamr_unify_client.Client`
    :param api_path: API path used to access this collection.
        E.g. ``"projects/1/attributeConfigurations"``
    :type api_path: str
    """

    def by_resource_id(self, resource_id):
        """Retrieve an attribute configuration by resource ID.

        :param resource_id: The resource ID.
        :type resource_id: str
        :returns: The specified attribute configuration.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfiguration`
        """
        return super().by_resource_id(self.api_path, resource_id)

    def by_relative_id(self, relative_id):
        """Retrieve an attribute configuration by relative ID.

       :param relative_id: The relative ID.
       :type relative_id: str
       :returns: The specified attribute configuration.
       :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfiguration`
       """
        return super().by_relative_id(AttributeConfiguration, relative_id)

    def by_external_id(self, external_id):
        """Retrieve an attribute configuration by external ID.

        Since attributes do not have external IDs, this method is not supported and will
        raise a :class:`NotImplementedError` .

        :param external_id: The external ID.
        :type external_id: str
        :returns: The specified attribute, if found.
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfiguration`
        :raises KeyError: If no attribute with the specified external_id is found
        :raises LookupError: If multiple attributes with the specified external_id are found
        :raises NotImplementedError: AttributeConfiguration does not support external_id
        """
        raise NotImplementedError("AttributeConfiguration does not support external_id")

    def stream(self):
        """Stream attribute configurations in this collection. Implicitly called when iterating
         over this collection.

         :returns: Stream of attribute configurations.
         :rtype: Python generator yielding :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfiguration`

         Usage:
             >>> for attributeConfiguration in collection.stream(): # explicit
             >>>     do_stuff(attributeConfiguration)
             >>> for attributeConfiguration in collection: # implicit
             >>>     do_stuff(attributeConfiguration)
         """

        return super().stream(AttributeConfiguration)

    def create(self, creation_spec):
        """Create an Attribute configuration in this collection

        :param creation_spec: Attribute configuration creation specification should be formatted as specified in the
            `Public Docs for adding an AttributeConfiguration <https://docs.tamr.com/reference#create-attribute-configurations>`_.
        :type creation_spec: dict[str, str]
        :returns: The created Attribute configuration
        :rtype: :class:`~tamr_unify_client.project.attribute_configuration.resource.AttributeConfiguration`
        """
        data = self.client.post(self.api_path, json=creation_spec).successful().json()
        return AttributeConfiguration.from_json(self.client, data)



import json

from tamr_unify_client.base_resource import BaseResource


class BinningModel(BaseResource):
    """ A binning model object."""

    @classmethod
    def from_json(cls, client, resource_json, api_path=None):
        return super().from_data(client, resource_json, api_path)

    def records(self):
        """Stream this object's records as Python dictionaries.

        :return: Stream of records.
        :rtype: Python generator yielding :py:class:`dict`
        """
        with self.client.get(self.api_path + "/records", stream=True) as response:
            for line in response.iter_lines():
                yield json.loads(line)

    def update_records(self, records):
        """Send a batch of record creations/updates/deletions to this dataset.

        :param records: Each record should be formatted as specified in the `Public Docs for Dataset updates <https://docs.tamr.com/reference#modify-a-datasets-records>`_.
        :type records: iterable[dict]
        :returns: JSON response body from server.
        :rtype: :py:class:`dict`
        """

        def _stringify_updates(updates):
            for update in updates:
                yield json.dumps(update).encode("utf-8")

        return (
            self.client.post(
                self.api_path + "/records",
                headers={"Content-Encoding": "utf-8"},
                data=_stringify_updates(records),
            )
            .successful()
            .json()
        )

    def __repr__(self):
        return (
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}("
            f"api_path={self.api_path})"
        )



"""
Convenient functionality for interacting with pandas DataFrames.
"""

import json
import os
from typing import Optional, TYPE_CHECKING

import requests

from tamr_client import attribute, dataset, primary_key
from tamr_client._types import Dataset, Instance, JsonDict, Session
from tamr_client.dataset import record
from tamr_client.exception import TamrClientException

BUILDING_DOCS = os.environ.get("TAMR_CLIENT_DOCS") == "1"
if TYPE_CHECKING or BUILDING_DOCS:
    import pandas as pd


class CreationFailure(TamrClientException):
    """Raised when a dataset could not be created from a pandas DataFrame
    """

    pass


def upsert(
    session: Session,
    dataset: Dataset,
    df: "pd.DataFrame",
    *,
    primary_key_name: Optional[str] = None,
) -> JsonDict:
    """Upserts a record for each row of `df` with attributes for each column in `df`.

    Args:
        dataset: Dataset to receive record updates
        df: The DataFrame containing records to be upserted
        primary_key_name: The primary key of the dataset.  Must be a column of `df`. By default the
            key_attribute_name of dataset

    Returns:
        JSON response body from the server

    Raises:
        requests.HTTPError: If an HTTP error is encountered
        primary_key.NotFound: If `primary_key_name` is not a column in `df` or the index of `df`
        ValueError: If `primary_key_name` matches both a column in `df` and the index of `df`
    """
    if primary_key_name is None:
        primary_key_name = dataset.key_attribute_names[0]

    # preconditions
    _check_primary_key(df, primary_key_name)

    # promote primary key column to index
    if primary_key_name in df.columns:
        df = df.set_index(primary_key_name)

    # serialize records via to_json to handle `np.nan` values
    serialized_records = ((pk, row.to_json()) for pk, row in df.iterrows())
    records = (
        {primary_key_name: pk, **json.loads(row)} for pk, row in serialized_records
    )
    return record.upsert(session, dataset, records, primary_key_name=primary_key_name)


def create(
    session: Session,
    instance: Instance,
    df: "pd.DataFrame",
    *,
    name: str,
    primary_key_name: Optional[str] = None,
    description: Optional[str] = None,
    external_id: Optional[str] = None,
) -> Dataset:
    """Create a dataset in Tamr from the DataFrame `df` and creates a record from each row

    All attributes other than the primary key are created as the default type array(string)

    Args:
        instance: Tamr instance
        df: The DataFrame containing records to be upserted
        name: Dataset name
        primary_key_name: The primary key of the dataset. Must be a column of `df`. By default the
            name of the index of `df`
        description: Dataset description
        external_id: External ID of the dataset

    Returns:
        Dataset created in Tamr

    Raises:
        dataset.AlreadyExists: If a dataset with these specifications already exists.
        requests.HTTPError: If any other HTTP error is encountered.
        primary_key.NotFound: If `primary_key_name` is not a column in `df` or the index of `df`
        ValueError: If `primary_key_name` matches both a column in `df` and the index of `df`
    """
    # preconditions
    if primary_key_name is None:
        if df.index.name is not None:
            primary_key_name = df.index.name
        else:
            raise primary_key.NotFound(
                "No primary key was specified and DataFrame index is unnamed"
            )
    _check_primary_key(df, primary_key_name)

    # dataset creation
    try:
        ds = dataset.create(
            session,
            instance,
            name=name,
            key_attribute_names=(primary_key_name,),
            description=description,
            external_id=external_id,
        )
    except (TamrClientException, requests.HTTPError) as e:
        raise CreationFailure(f"Dataset was not created: {e}")

    # attribute creation
    for col in df.columns:
        if col == primary_key_name:
            # this attribute already exists as a key attribute
            continue
        try:
            attribute.create(session, ds, name=col, is_nullable=True)
        except (TamrClientException, requests.HTTPError) as e:
            _handle_creation_failure(session, ds, f"An attribute was not created: {e}")

    # record creation
    try:
        response = upsert(session, ds, df, primary_key_name=primary_key_name)
        if not response["allCommandsSucceeded"]:
            _handle_creation_failure(session, ds, "Some records had validation errors")
    except (TamrClientException, requests.HTTPError) as e:
        _handle_creation_failure(session, ds, f"Record could not be created: {e}")

    # Get Dataset from server
    return dataset._dataset._by_url(session, ds.url)


def _handle_creation_failure(session: Session, stub_dataset: Dataset, error: str):
    """Attempt to make `dataframe.create` atomic by deleting the created dataset in the event of
    later failure.

    However, this does not guarantee atomicity: if the request to delete the dataset fails, it will
    not retry.

    Args:
        stub_dataset: The created dataset to delete
        error: The error that caused dataset creation to fail
    """
    try:
        dataset.delete(session, stub_dataset)
    except requests.HTTPError:
        raise CreationFailure(
            f"Created dataset did not delete after an earlier error: {error}"
        )
    raise CreationFailure(error)


def _check_primary_key(df: "pd.DataFrame", primary_key_name: str):
    """Check if the primary key name uniquely identifies a column or index of the DataFrame

    Args:
        df: The DataFrame to inspect
        primary_key_name: The index or column name to be used as the primary key

    Raises:
        primary_key.Ambiguous: If the primary key name matches both the index and a column
        primary_key.NotFound: If the primary key name does not match the index or any column
    """
    if primary_key_name in df.columns and primary_key_name == df.index.name:
        raise primary_key.Ambiguous(
            f"Index {primary_key_name} has the same name as column {primary_key_name}"
        )
    elif primary_key_name not in df.columns and primary_key_name != df.index.name:
        raise primary_key.NotFound(
            f"Primary key: {primary_key_name} is not DataFrame index name: {df.index.name} or in"
            f" DataFrame column names: {df.columns}"
        )
